/*
//  pgipfauth - Postgres IPF Authorization Helper
//  IPFAuthCache
//
//  Authorized IP address caching.
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: IPFAuthCache.c 122 2009-03-04 17:03:48Z frey $
*/

#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>

#include "IPFAuthCache.h"


/*!
  @constant IPFAuthCacheTypeStrings
  
  Textual descriptions of the cache types for the sake of being human-readable.
*/
const char*   IPFAuthCacheTypeStrings[3] = {
                  "oldest-first" ,
                  "newest-first" ,
                  "stateful"
                };


#ifndef IPFAUTHCACHE_MINIMUM_TTL
/*!
  @defined IPFAUTHCACHE_MINIMUM_TTL
  
  Minimum cache lifetime for IP addresses.  Defaults to 5 minutes, but can
  be overridden by using a -DIPFAUTHCACHE_MINIMUM_TTL=### in your CFLAGS.
*/
#define IPFAUTHCACHE_MINIMUM_TTL 5
#endif


#ifndef IPFAUTHCACHE_MINIMUM_SIZE
/*!
  @defined IPFAUTHCACHE_MINIMUM_SIZE
  
  Minimum number of cache slots to allocate.  Defaults to 256 slots, but can
  be overridden by using a -DIPFAUTHCACHE_MINIMUM_SIZE=### in your CFLAGS.
*/
#define IPFAUTHCACHE_MINIMUM_SIZE 4
#endif


#ifndef IPFAUTHCACHE_MINIMUM_GROW
/*!
  @defined IPFAUTHCACHE_MINIMUM_GROW
  
  Minimum number of new slots that should be added to an adaptive cache when
  the critical miss rate is reached.  Defaults to 128 slots, but can be
  overridden by using a -DIPFAUTHCACHE_MINIMUM_GROW=### in your CFLAGS.
*/
#define IPFAUTHCACHE_MINIMUM_GROW 2
#endif


#ifndef IPFAUTHCACHE_CRITICAL_PCT
/*!
  @defined IPFAUTHCACHE_CRITICAL_PCT
  
  Default percentage of misses that will trigger a growth of an adaptive
  cache when full.  Defaults to 5% (0.05), but can be overridden by using
  a -DIPFAUTHCACHE_CRITICAL_PCT=0.### in your CFLAGS.
*/
#define IPFAUTHCACHE_CRITICAL_PCT 0.05
#endif


/*!
  @enum Cache entry flags
  
  Flags that control cache nodes.
*/
enum {
  kIPFAuthCacheEntry_InUse      = 1 << 0,
  kIPFAuthCacheEntry_IPIsAuthzd = 1 << 1
};


/*!
  @typedef IPFAuthCacheEntry
  
  The data portion of a node in the cache queue.
*/
typedef struct __IPFAuthCacheEntry {
  unsigned int                flags;
  struct in_addr              ipAddress;
  int                         ipPort;
  time_t                      expires;
} IPFAuthCacheEntry;


/*!
  @typedef IPFAuthCacheQ
  
  A node in the cache queue.  The queue is a circular doubly-linked
  list.
*/
typedef struct __IPFAuthCacheQ {
  IPFAuthCacheEntry           data;
  struct __IPFAuthCacheQ*     fLink;
  struct __IPFAuthCacheQ*     bLink;
} IPFAuthCacheQ;


/*!
  @function __IPFAuthCacheQ_Print
  
  Print a description of a cache queue node.
*/
void
__IPFAuthCacheQ_Print(
  IPFAuthCacheQ*    node
)
{
  if ( node->data.flags & kIPFAuthCacheEntry_InUse ) {
    fprintf(
        stderr,
        "%p <%p|%p> { flags: %02X | ip: %s:%d | expires: %ld }",
        node,
        node->bLink,
        node->fLink,
        node->data.flags,
        inet_ntoa(node->data.ipAddress),
        node->data.ipPort,
        (long int)(node->data.expires - time(NULL))
      );
  } else {
    fprintf(
        stderr,
        "%p <%p|%p> { flags: %02X | empty }",
        node,
        node->bLink,
        node->fLink,
        node->data.flags
      );
  }
}


/*!
  @typedef IPFAuthCacheLookupCallback
  
  Pointer to a function that does cache lookup.
*/
typedef IPFAuthCacheQ* (*IPFAuthCacheLookupCallback)(struct __IPFAuthCache* aCache,struct in_addr* ipAddress,int ipPort);


/*!
  @enum Cache flags
  
  Internally-used flags that alter cache behavior.
*/
enum {
  kIPFAuthCacheIsAdaptive       = 1 << 0,
  kIPFAuthCacheDisableAdaptive  = 1 << 1,
  kIPFAuthCacheHonorIPPort      = 1 << 2,
  
  kIPFAuthCacheIsFull           = 1 << 16
};


/*!
  @typedef IPFAuthCache
  
  The authorization cache pseudo-object data structure.  This struct
  is private; the API only communicates back and forth via opaque
  pointers to such a struct (type IPFAuthCacheRef).
*/
typedef struct __IPFAuthCache {
  unsigned int                refCount;
  unsigned int                flags;
  
  IPFAuthCacheType            type;
  IPFAuthCacheLookupCallback  lookupProc;
  
  unsigned int                qSize;
  IPFAuthCacheQ*              qTrueHead;
  IPFAuthCacheQ*              qHead;
  IPFAuthCacheQ*              qTail;
  IPFAuthCacheQ*              qLast;
  time_t                      ttl;
  
  unsigned int                growBy;
  float                       missFraction;
  unsigned int                miss,total;
} IPFAuthCache;


/*!
  @function __IPFAuthCache_InitQueue
  
  Allocates and initializes an authorization cache queue of
  arbitrary length and sets the applicable fields in a
  IPFAuthCache pseudo-object struct.
  
  If aCache already had a queue attached to it, that queue
  should have been destroyed before this function is
  called.
*/
int
__IPFAuthCache_InitQueue(
  IPFAuthCache*       aCache,
  unsigned int        qSize
)
{
  IPFAuthCacheQ*      head = calloc(qSize,sizeof(IPFAuthCacheQ));
  
  if ( head ) {
    /* Initialize the cache size and pointers: */
    aCache->qSize = qSize;
    aCache->qTrueHead = aCache->qLast = aCache->qHead = aCache->qTail = head;
    
    /* Set the links on all but the last queue node... */
    while ( --qSize ) {
      head->fLink = head + 1;
      (head + 1)->bLink = head;
      head++;
    }
    
    /* ...and link the final node back to the head node. */
    head->fLink = aCache->qHead;
    aCache->qHead->bLink = head;
    
    return 1;
  }
  return 0;
}


/*!
  @function __IPFAuthCache_ExtendQueue
  
  Attempt to increase the size of the queue by adding dQSize slots between
  the original head and tail nodes.  Returns non-zero on success.
  
  Preserves the head, tail, and last pointers (relative to the new memory
  block realloc'ed).
*/
int
__IPFAuthCache_ExtendQueue(
  IPFAuthCache*       aCache,
  unsigned int        dQSize
)
{
  IPFAuthCacheQ*      newQ;
  unsigned int        newSize;
  
  /* Ha ha, like we'll ever have over UINT_MAX cache nodes anyway! */
  if ( (UINT_MAX - dQSize) < aCache->qSize ) {
    newSize = UINT_MAX;
    dQSize = UINT_MAX - aCache->qSize;
  } else {
    newSize = aCache->qSize + dQSize;
  }
  
  /* Try a realloc: */
  newQ = realloc(aCache->qTrueHead,sizeof(IPFAuthCacheQ) * newSize);
  if ( newQ ) {
    unsigned int      oldHead = aCache->qHead - aCache->qTrueHead;
    unsigned int      oldTail = aCache->qTail - aCache->qTrueHead;
    unsigned int      oldLast = aCache->qLast - aCache->qTrueHead;
    
    /* Zero the new chunk: */
    memset(newQ + aCache->qSize, 0, sizeof(IPFAuthCacheQ) * dQSize);
    
    /* Reset all pointers relative to the newly-allocated queue: */
    aCache->qSize = newSize;
    aCache->qTrueHead = newQ;
    aCache->qHead = newQ + oldHead;
    aCache->qTail = newQ + oldTail;
    aCache->qLast = newQ + oldLast;
    
    /* Turn-off the "full" bit: */
    aCache->flags &= ~kIPFAuthCacheIsFull;
    
    /* Reset the hit/miss counts: */
    aCache->miss = aCache->total = 0;
    
    /* Re-link all the nodes; set the links on all but the last queue node... */
    while ( --newSize ) {
      newQ->fLink = newQ + 1;
      (newQ + 1)->bLink = newQ;
      newQ++;
    }
    
    /* ...and link the final node back to the head node. */
    newQ->fLink = aCache->qTrueHead;
    aCache->qTrueHead->bLink = newQ;
    
    return 1;
  }
  /* Disable adaptive and DON'T allow it to come back: */
  IPFAuthCacheSetIsAdaptive((IPFAuthCacheRef)aCache,0,0,0.0f);
  aCache->flags |= kIPFAuthCacheDisableAdaptive;
  return 0;
}


/*!
  @function __IPFAuthCache_Reset
  
  Nulls-out all cache queue nodes and resets the node pointers in
  aCache.
*/
void
__IPFAuthCache_Reset(
  IPFAuthCache*       aCache
)
{
  IPFAuthCacheQ*      node = aCache->qHead;
  
  /* Tail pointer will be back at the head node: */
  aCache->qTail = aCache->qHead;
  
  /* Zero-out the data fields on all the queue nodes: */
  do {
    memset(&node->data, 0, sizeof(IPFAuthCacheEntry));
    node = node->fLink;
  } while ( node != aCache->qHead );
}


/*!
  @function __IPFAuthCache_LookupOldestFirst
  
  Locate and return the cache queue node that contains the specified IP
  address.  The search proceeds from the oldest node to the newest.
*/
IPFAuthCacheQ*
__IPFAuthCache_LookupOldestFirst(
  IPFAuthCache*       aCache,
  struct in_addr*     ipAddress,
  int                 ipPort
)
{
  IPFAuthCacheQ*      node = aCache->qHead;
  
  do {
    if ( ! (node->data.flags & kIPFAuthCacheEntry_InUse) ) {
      /* The rest of the nodes (including this one) are not used. */
      break;
    }
    if ( (node->data.ipPort == ipPort) && (memcmp(ipAddress,&node->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
      /* Found it! */
      aCache->qLast = node;
      return node;
    }
    node = node->fLink;
  } while ( node != aCache->qHead );
  
  /* Not found */
  return NULL;
}


/*!
  @function __IPFAuthCache_LookupNewestFirst
  
  Locate and return the cache queue node that contains the specified IP
  address.  The search proceeds from the newest node to the oldest.
*/
IPFAuthCacheQ*
__IPFAuthCache_LookupNewestFirst(
  IPFAuthCache*       aCache,
  struct in_addr*     ipAddress,
  int                 ipPort
)
{
  IPFAuthCacheQ*      node = aCache->qTail;
  
  /* Nothing in the cache: */
  if ( (node == aCache->qHead) && ( ! (node->data.flags & kIPFAuthCacheEntry_InUse) ) ) {
    return NULL;
  }
  
  do {
    node = node->bLink;
    
    if ( ! (node->data.flags & kIPFAuthCacheEntry_InUse) ) {
      /*  We should honestly never get here: */
      continue;
    }
    if ( (node->data.ipPort == ipPort) && (memcmp(ipAddress,&node->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
      /* Found it! */
      aCache->qLast = node;
      return node;
    }
  } while ( node != aCache->qHead );
  
  /* Not found */
  return NULL;
}


/*!
  @function __IPFAuthCache_LookupStateful
  
  Locate and return the cache queue node that contains the specified IP
  address.  The search proceeds from the last IP address that was either
  looked-up or added to the cache.
*/
IPFAuthCacheQ*
__IPFAuthCache_LookupStateful(
  IPFAuthCache*       aCache,
  struct in_addr*     ipAddress,
  int                 ipPort
)
{
  IPFAuthCacheQ*      fNode = aCache->qLast;
  IPFAuthCacheQ*      bNode = fNode;
  int                 contFrwd = 0,contBkwd =0;
  
  /* The state points to something that's no longer cached; fall-back on
     newest-first: */
#ifdef IPFAUTHCACHE_DEBUG
  printf("I %p\n",fNode);
#endif
  if ( ! (fNode->data.flags & kIPFAuthCacheEntry_InUse) ) {
    return __IPFAuthCache_LookupNewestFirst(aCache, ipAddress, ipPort);
  } else if ( (fNode->data.ipPort == ipPort) && (memcmp(ipAddress,&fNode->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
    /* The last one was the one we wanted! */
    aCache->qLast = fNode;
    return fNode;
  }
  
  do {
    bNode = bNode->bLink;
#ifdef IPFAUTHCACHE_DEBUG
    printf("B ");__IPFAuthCacheQ_Print(bNode);printf("\n");
#endif
    if ( bNode == fNode ) {
      break;
    }
    if ( bNode == aCache->qLast ) {
      contFrwd = 1;
      break;
    }
    if ( (bNode->data.flags & kIPFAuthCacheEntry_InUse) ) {
      if ( (bNode->data.ipPort == ipPort) && (memcmp(ipAddress,&bNode->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
        /* Found it! */
        aCache->qLast = bNode;
        return bNode;
      }
    } else {
      contFrwd = 1;
      break;
    }
    
    fNode = fNode->fLink;
#ifdef IPFAUTHCACHE_DEBUG
    printf("F ");__IPFAuthCacheQ_Print(fNode);printf("\n");
#endif
    if ( fNode == bNode ) {
      break;
    }
    if ( fNode == aCache->qLast ) {
      contBkwd = 1;
      break;
    }
    if ( (fNode->data.flags & kIPFAuthCacheEntry_InUse) ) {
      if ( (fNode->data.ipPort == ipPort) && (memcmp(ipAddress,&fNode->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
        /* Found it! */
        aCache->qLast = fNode;
        return fNode;
      }
    } else {
      contBkwd = 1;
      break;
    }
  } while ( 1 );
  
  if ( contFrwd ) {
    fNode = fNode->fLink;
    while ( (fNode != bNode) && (fNode != aCache->qLast) ) {
#ifdef IPFAUTHCACHE_DEBUG
      printf("F ");__IPFAuthCacheQ_Print(fNode);printf("\n");
#endif
      if ( (fNode->data.flags & kIPFAuthCacheEntry_InUse) ) {
        if ( (fNode->data.ipPort == ipPort) && (memcmp(ipAddress,&fNode->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
          /* Found it! */
          aCache->qLast = fNode;
          return fNode;
        }
      } else {
        break;
      }
      fNode = fNode->fLink;
    }
  } else if ( contBkwd ) {
    bNode = bNode->bLink;
    while ( (bNode != fNode) && (bNode != aCache->qLast) ) {
#ifdef IPFAUTHCACHE_DEBUG
      printf("B ");__IPFAuthCacheQ_Print(bNode);printf("\n");
#endif
      if ( (bNode->data.flags & kIPFAuthCacheEntry_InUse) ) {
        if ( (bNode->data.ipPort == ipPort) && (memcmp(ipAddress,&bNode->data.ipAddress,sizeof(struct in_addr)) == 0) ) {
          /* Found it! */
          aCache->qLast = bNode;
          return bNode;
        }
      } else {
        break;
      }
      bNode = bNode->bLink;
    }
  }
  
  /* Not found */
  return NULL;
}


/*!
  @function __IPFAuthCache_AddIPAddress
  
  Add the given ipAddress to the cache queue.  Always uses the node pointed-at
  by the cache's tail node pointer.  Eventually the tail will wrap back around
  to the head, at which point we'll begin pushing both the head and tail ahead
  with each new address added to the cache.  The oldest cache record is always
  indicated by the head node pointer, so we in essence merely begin overwriting
  the oldest cached address that hasn't yet timed-out.
*/
void
__IPFAuthCache_AddIPAddress(
  IPFAuthCache*         aCache,
  struct in_addr*       ipAddress,
  int                   ipPort,
  IPFAuthDisposition    isAuthorized
)
{
  IPFAuthCacheQ*      node = aCache->qTail;
  unsigned int        wasInUse = (node->data.flags & kIPFAuthCacheEntry_InUse);
  
  /* Use the tail node: */
  memcpy(&node->data.ipAddress,ipAddress,sizeof(struct in_addr));
  node->data.ipPort = ipPort;
  node->data.flags = kIPFAuthCacheEntry_InUse;
  switch ( isAuthorized ) {
    case kIPFAuthDisposition_Pass:
      node->data.flags |= kIPFAuthCacheEntry_IPIsAuthzd;
      break;
  }
  node->data.expires = time(NULL) + aCache->ttl;
  aCache->qLast = node;
  
  /* If the tail and head pointers are the same AND the tail node
     was already in-use, then we have to push the head AND the tail
     pointers forward; otherwise, only push the tail pointer: */
  if ( (aCache->qHead == aCache->qTail) && wasInUse ) {
    aCache->qHead = node->fLink;
    aCache->flags |= kIPFAuthCacheIsFull;
  }
  aCache->qTail = node->fLink;
}


/*!
  @function __IPFAuthCache_Cleanup
  
  Purge all expired cache entries.
*/
void
__IPFAuthCache_Cleanup(
  IPFAuthCache*       aCache
)
{
  IPFAuthCacheQ*      node = aCache->qHead;
  time_t              now = time(NULL);
  
  while ( (node->data.flags & kIPFAuthCacheEntry_InUse) && (node->data.expires <= now) ) {
    node->data.flags = 0;
    aCache->qHead = node = node->fLink;
  }
}


/*!
  @function __IPFAuthCache_InvalidateIP
  
  Invalidate any cache records for a specific IP from the cache.
*/
void
__IPFAuthCache_InvalidateIP(
  IPFAuthCache*       aCache,
  struct in_addr*     anIP
)
{
  IPFAuthCacheQ*      node = aCache->qHead;
  
  while ( node->data.flags & kIPFAuthCacheEntry_InUse ) {
    if ( memcmp(&node->data.ipAddress, anIP, sizeof(struct in_addr)) == 0 ) {
      /* Ha...set the IP to null and let it eventually expire...it's the easy way out. */
      node->data.flags = kIPFAuthCacheEntry_InUse;
      memset(&node->data.ipAddress, 0, sizeof(struct in_addr));
      node->data.ipPort = 0;
    }
    node = node->fLink;
    
    if ( node == aCache->qHead )
      break;
  }
}

/*   P U B L I C   F U N C T I O N S   */


IPFAuthCacheRef
IPFAuthCacheCreate(
  IPFAuthCacheType  type,
  unsigned int      size,
  time_t            ttl
)
{
  IPFAuthCache*   newCache = calloc(1,sizeof(IPFAuthCache));
  
  if ( newCache ) {
    
    if ( size < IPFAUTHCACHE_MINIMUM_SIZE )
      size = IPFAUTHCACHE_MINIMUM_SIZE;
      
    if ( __IPFAuthCache_InitQueue(newCache,size) ) {
      newCache->refCount    = 1;
      newCache->ttl         = ( ttl >= IPFAUTHCACHE_MINIMUM_TTL ? ttl : IPFAUTHCACHE_MINIMUM_TTL );
      
      IPFAuthCacheSetType((IPFAuthCacheRef)newCache, type);
      IPFAuthCacheSetHonorIPPort((IPFAuthCacheRef)newCache, 1);
    } else {
      free(newCache);
      newCache = NULL;
    }
  }
  return newCache;
}


IPFAuthCacheRef
IPFAuthCacheRetain(
  IPFAuthCacheRef anAuthCache
)
{
  ((IPFAuthCache*)anAuthCache)->refCount++;
  return anAuthCache;
}


void
IPFAuthCacheRelease(
  IPFAuthCacheRef anAuthCache
)
{
  if ( --((IPFAuthCache*)anAuthCache)->refCount == 0 ) {
    if ( anAuthCache->qTrueHead ) free(anAuthCache->qTrueHead);
    free((IPFAuthCache*)anAuthCache);
  }
}


IPFAuthCacheType
IPFAuthCacheGetType(
  IPFAuthCacheRef anAuthCache
)
{
  return anAuthCache->type;
}


void
IPFAuthCacheSetType(
  IPFAuthCacheRef   anAuthCache,
  IPFAuthCacheType  type
)
{
  if ( type != anAuthCache->type ) {
    switch ( type ) {
      
      case kIPFAuthCacheTypeOldestFirst:
        ((IPFAuthCache*)anAuthCache)->lookupProc = __IPFAuthCache_LookupOldestFirst;
        break;
      
      case kIPFAuthCacheTypeStateful:
        ((IPFAuthCache*)anAuthCache)->lookupProc = __IPFAuthCache_LookupStateful;
        break;
      
      default:
        type = kIPFAuthCacheTypeNewestFirst;
      case kIPFAuthCacheTypeNewestFirst:
        ((IPFAuthCache*)anAuthCache)->lookupProc = __IPFAuthCache_LookupNewestFirst;
        break;
        
    }
    ((IPFAuthCache*)anAuthCache)->type = type;
  }
}


unsigned int
IPFAuthCacheGetSize(
  IPFAuthCacheRef anAuthCache
)
{
  return anAuthCache->qSize;
}


time_t
IPFAuthCacheGetTTL(
  IPFAuthCacheRef anAuthCache
)
{
  return anAuthCache->ttl;
}


void
IPFAuthCacheSetTTL(
  IPFAuthCacheRef anAuthCache,
  time_t          ttl
)
{
  if ( ttl >= IPFAUTHCACHE_MINIMUM_TTL ) {
    ((IPFAuthCache*)anAuthCache)->ttl = ttl;
  }
}


int
IPFAuthCacheIsAdaptive(
  IPFAuthCacheRef anAuthCache
)
{
  return ( (anAuthCache->flags & kIPFAuthCacheIsAdaptive) != 0 );
}


void
IPFAuthCacheSetIsAdaptive(
  IPFAuthCacheRef anAuthCache,
  int             isAdaptive,
  unsigned int    growBy,
  float           atMissFraction
)
{
  if ( (anAuthCache->flags & kIPFAuthCacheDisableAdaptive) == 0 ) {
    if ( isAdaptive ) {
      ((IPFAuthCache*)anAuthCache)->flags |= kIPFAuthCacheIsAdaptive;
      ((IPFAuthCache*)anAuthCache)->growBy = ( growBy > IPFAUTHCACHE_MINIMUM_GROW ? growBy : IPFAUTHCACHE_MINIMUM_GROW );
      
      if ( (atMissFraction < 0.05f) || (atMissFraction > 0.95f) ) {
        ((IPFAuthCache*)anAuthCache)->missFraction = IPFAUTHCACHE_CRITICAL_PCT;
      } else {
        ((IPFAuthCache*)anAuthCache)->missFraction = atMissFraction;
      }
    } else {
      ((IPFAuthCache*)anAuthCache)->flags &= ~kIPFAuthCacheIsAdaptive;
    }
    
    /* Reset statistical counters: */
    ((IPFAuthCache*)anAuthCache)->total = ((IPFAuthCache*)anAuthCache)->miss = 0;
  }
}


int
IPFAuthCacheGetHonorIPPort(
  IPFAuthCacheRef   anAuthCache
)
{
  return ( (anAuthCache->flags & kIPFAuthCacheHonorIPPort) != 0 );
}


void
IPFAuthCacheSetHonorIPPort(
  IPFAuthCacheRef   anAuthCache,
  int               honorIPPort
)
{
  if ( ( honorIPPort != 0 ) != ( (anAuthCache->flags & kIPFAuthCacheHonorIPPort) != 0 ) ) {
    ((IPFAuthCache*)anAuthCache)->flags ^= kIPFAuthCacheHonorIPPort;
    IPFAuthCachePurgeAll(anAuthCache);
  }
}


void
IPFAuthCachePurgeExpired(
  IPFAuthCacheRef anAuthCache
)
{
  __IPFAuthCache_Cleanup((IPFAuthCache*)anAuthCache);
}



void
IPFAuthCachePurgeAll(
  IPFAuthCacheRef anAuthCache
)
{
  __IPFAuthCache_Reset((IPFAuthCache*)anAuthCache);
}


void
IPFAuthCacheInvalidateIP(
  IPFAuthCacheRef anAuthCache,
  struct in_addr* ipAddress
)
{
  __IPFAuthCache_InvalidateIP((IPFAuthCache*)anAuthCache, ipAddress);
}


int
IPFAuthCacheIsAuthorized(
  IPFAuthCacheRef       anAuthCache,
  struct in_addr*       ipAddress,
  int                   ipPort,
  IPFAuthDisposition*   isAuthorized
)
{
  IPFAuthCacheQ*        node = NULL;
  
  if ( ! (anAuthCache->flags & kIPFAuthCacheHonorIPPort) )
    ipPort = 0;
  
  node = (anAuthCache->lookupProc)((IPFAuthCache*)anAuthCache, ipAddress, ipPort);
  
  if ( anAuthCache->flags & kIPFAuthCacheIsAdaptive ) {
    /* Adaptive, we've gotta update our stats: */
    if ( node && (node->data.expires > time(NULL)) ) {
      /* Update the total counter */
      if ( ++((IPFAuthCache*)anAuthCache)->total == UINT_MAX ) {
        if ( anAuthCache->miss ) {
          /* Drop the total as far as it can go by setting the miss count to
             one and dividing the total by the old miss count: */
          ((IPFAuthCache*)anAuthCache)->total /= anAuthCache->miss;
          ((IPFAuthCache*)anAuthCache)->miss = 1;
        } else
          ((IPFAuthCache*)anAuthCache)->total = 0;
      }
      *isAuthorized = ( (node->data.flags & kIPFAuthCacheEntry_IPIsAuthzd) != 0 ? kIPFAuthDisposition_Pass : kIPFAuthDisposition_Block);
      return 1;
    }
    /* It was a miss; update the miss and total counters... */
    if ( ++((IPFAuthCache*)anAuthCache)->miss == UINT_MAX ) {
      ((IPFAuthCache*)anAuthCache)->miss = ((IPFAuthCache*)anAuthCache)->total = 1;
    } else if ( ++((IPFAuthCache*)anAuthCache)->total == UINT_MAX ) {
      /* Drop the total as far as it can go by setting the miss count to
         one and dividing the total by the old miss count: */
      ((IPFAuthCache*)anAuthCache)->total /= anAuthCache->miss;
      ((IPFAuthCache*)anAuthCache)->miss = 1;
    }
      
    /* Now let's check the miss percentage if we're out of nodes: */
    if ( (anAuthCache->flags & kIPFAuthCacheIsFull) && ( ((float)anAuthCache->miss / (float)anAuthCache->total) >= anAuthCache->missFraction ) ) {
      /* Attempt to grow the cache; the called routine will disable the adapative mechanism
         if it fails to grow: */
      __IPFAuthCache_ExtendQueue((IPFAuthCache*)anAuthCache,anAuthCache->growBy);
    }
    return 0;
  }
  
  /* Not adaptive, just quick-as-can-be behavior: */
  if ( node && (node->data.expires > time(NULL)) ) {
    *isAuthorized = ( (node->data.flags & kIPFAuthCacheEntry_IPIsAuthzd) != 0 ? kIPFAuthDisposition_Pass : kIPFAuthDisposition_Block);
    return 1;
  }
  return 0;
}


void
IPFAuthCacheAuthorize(
  IPFAuthCacheRef       anAuthCache,
  struct in_addr*       ipAddress,
  int                   ipPort,
  IPFAuthDisposition    isAuthorized
)
{
  IPFAuthCacheQ*        node = NULL;
  
  if ( ! (anAuthCache->flags & kIPFAuthCacheHonorIPPort) )
    ipPort = 0;
  
  node = (anAuthCache->lookupProc)((IPFAuthCache*)anAuthCache, ipAddress, ipPort);
  
  if ( node ) {
    switch ( isAuthorized ) {
      
      case kIPFAuthDisposition_Pass: {
        if ( (node->data.flags & kIPFAuthCacheEntry_IPIsAuthzd) == 0 )
          node->data.flags |= kIPFAuthCacheEntry_IPIsAuthzd;
        break;
      }
      
      case kIPFAuthDisposition_Block: {
        if ( node->data.flags & kIPFAuthCacheEntry_IPIsAuthzd )
          node->data.flags &= ~kIPFAuthCacheEntry_IPIsAuthzd;
        break;
      }
      
    }
  } else
    __IPFAuthCache_AddIPAddress((IPFAuthCache*)anAuthCache, ipAddress, ipPort, isAuthorized);
}


void
IPFAuthCacheDebug(
  IPFAuthCacheRef anAuthCache
)
{
  IPFAuthCacheQ*        node = anAuthCache->qHead;
  
  fprintf(stderr,"     Cache type %s of size %u (flags = %08X) miss pct: %.2f\n     real head: %p, head: %p, tail: %p, last: %p\n",
        IPFAuthCacheTypeStrings[anAuthCache->type - kIPFAuthCacheTypeOldestFirst],
        anAuthCache->qSize,
        anAuthCache->flags,
        ( anAuthCache->total ? ((float)anAuthCache->miss / (float)anAuthCache->total) : 0.0f ),
        anAuthCache->qTrueHead,
        anAuthCache->qHead,
        anAuthCache->qTail,
        anAuthCache->qLast
      );
  
  do {
    fprintf(stderr,"%c%c%c  ",
        ( node == anAuthCache->qHead ? 'h' : ' ' ),
        ( node == anAuthCache->qTail ? 't' : ' ' ),
        ( node == anAuthCache->qLast ? 'l' : ' ' )
      );
    __IPFAuthCacheQ_Print(node);
    fprintf(stderr,"\n");
    
    node = node->fLink;
  } while (node != anAuthCache->qHead);
}


void
IPFAuthCacheInfo(
  IPFAuthCacheRef anAuthCache
)
{
  printf("  cache summary {\n"
         "    type: %s, %s\n"
         "    ttl: %lu\n"
         "    size: {\n"
         "      nodes: %u\n",
      IPFAuthCacheTypeStrings[anAuthCache->type - kIPFAuthCacheTypeOldestFirst],
      ( IPFAuthCacheGetHonorIPPort(anAuthCache) ? "IP-port-specific" : "any IP port" ),
      (long unsigned int)anAuthCache->ttl,
      anAuthCache->qSize
    );
  if ( anAuthCache->flags & kIPFAuthCacheIsAdaptive ) {
    printf("      adaptive: yes\n"
           "      grow-by: %u\n"
           "      miss-percent: critical = %.1f%% ; current = %.1f%%\n",
        anAuthCache->growBy,
        anAuthCache->missFraction * 100.0f,
        ( anAuthCache->total ? ( (float)anAuthCache->miss / (float)anAuthCache->total ) * 100.0f : 0.0f )
      );
  }
  printf("    }\n"
         "    is-full: %s\n"
         "  }\n",
      ( anAuthCache->flags & kIPFAuthCacheIsFull ? "yes" : "no" )
    );
}


#ifdef HAS_LIBXML2
#include <libxml/parser.h>

const xmlChar* IPFAuthCache_Element_cache = BAD_CAST("cache");
const xmlChar* IPFAuthCache_Attribute_enabled = BAD_CAST("enabled");
const xmlChar* IPFAuthCache_Attribute_size = BAD_CAST("size");
const xmlChar* IPFAuthCache_Attribute_ttl = BAD_CAST("ttl");
const xmlChar* IPFAuthCache_Attribute_honorIPPort = BAD_CAST("honor-ip-port");

const xmlChar* IPFAuthCache_Element_search = BAD_CAST("search");
const xmlChar* IPFAuthCache_Attribute_method = BAD_CAST("method");
const xmlChar* IPFAuthCache_Attribute_method_default = BAD_CAST("default");
const xmlChar* IPFAuthCache_Attribute_method_newest = BAD_CAST("newest");
const xmlChar* IPFAuthCache_Attribute_method_oldest = BAD_CAST("oldest");
const xmlChar* IPFAuthCache_Attribute_method_stateful = BAD_CAST("stateful");

const xmlChar* IPFAuthCache_Element_adaptive = BAD_CAST("adaptive");
const xmlChar* IPFAuthCache_Attribute_growBy = BAD_CAST("grow-by");
const xmlChar* IPFAuthCache_Attribute_critFrac = BAD_CAST("critical-fraction");

/* Defined in pgipfauth.c */
int xmlCharToBool(const xmlChar* s);

IPFAuthCacheRef
IPFAuthCacheCreateWithXMLNode(
  xmlNodePtr      xmlNode
)
{
  IPFAuthCacheRef newCache = NULL;
  
  /* Look for our node: */
  while ( xmlNode ) {
    if ( xmlNode->type == XML_ELEMENT_NODE ) {
      if ( xmlStrcasecmp(xmlNode->name,IPFAuthCache_Element_cache) == 0 ) {
        break;
      }
    }
    xmlNode = xmlNode->next;
  }
  
  if ( xmlNode ) {
    xmlChar*    attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_enabled);
    
    if ( attrVal ) {
      if ( xmlCharToBool(attrVal) ) {
        unsigned int      size = 0,ttl = 0,growBy = 0;
        float             criticalFrac = 0.0;
        IPFAuthCacheType  type = kIPFAuthCacheTypeDefault;
        int               adaptive = 0;
        int               honorIPPort = 1;
        
        xmlFree(attrVal);
        
        /* Fetch the size attribute: */
        if ( (attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_size)) ) {
          size = strtol((const char*)attrVal,NULL,10);
          xmlFree(attrVal);
        }
        
        /* Fetch the ttl attribute: */
        if ( (attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_ttl)) ) {
          ttl = strtol((const char*)attrVal,NULL,10);
          xmlFree(attrVal);
        }
        
        /* Honor IP port? */
        if ( (attrVal = xmlGetProp(xmlNode, IPFAuthCache_Attribute_honorIPPort)) ) {
          honorIPPort = xmlCharToBool(attrVal);
          xmlFree(attrVal);
        }
        
        /* Check for the search and adaptive elements: */
        xmlNode = xmlNode->children;
        while ( xmlNode ) {
          if ( xmlNode->type == XML_ELEMENT_NODE ) {
            if ( xmlStrcasecmp(xmlNode->name,IPFAuthCache_Element_search) == 0 ) {
              /* Do we have a method attribute? */
              if ( attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_method) ) {
                if ( xmlStrcasecmp(attrVal,IPFAuthCache_Attribute_method_default) == 0 ) {
                  type = kIPFAuthCacheTypeDefault;
                } else
                if ( xmlStrcasecmp(attrVal,IPFAuthCache_Attribute_method_oldest) == 0 ) {
                  type = kIPFAuthCacheTypeOldestFirst;
                } else
                if ( xmlStrcasecmp(attrVal,IPFAuthCache_Attribute_method_newest) == 0 ) {
                  type = kIPFAuthCacheTypeNewestFirst;
                } else
                if ( xmlStrcasecmp(attrVal,IPFAuthCache_Attribute_method_stateful) == 0 ) {
                  type = kIPFAuthCacheTypeStateful;
                } else
                  fprintf(stderr,"WARNING: Invalid cache search method: %s\n",attrVal);
                xmlFree(attrVal);
              }
            } else
            if ( xmlStrcasecmp(xmlNode->name,IPFAuthCache_Element_adaptive) == 0 ) {
              /* Is the adaptive cache feature enabled? */
              if ( attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_enabled) ) {
                if ( xmlCharToBool(attrVal) ) {
                  adaptive = 1;
                  xmlFree(attrVal);
                  
                  /* Fetch the grow-by attribute: */
                  if ( (attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_growBy)) ) {
                    growBy = strtol((const char*)attrVal,NULL,10);
                    xmlFree(attrVal);
                  }
                  
                  /* Fetch the critical fraction attribute: */
                  if ( (attrVal = xmlGetProp(xmlNode,IPFAuthCache_Attribute_critFrac)) ) {
                    criticalFrac = strtof((const char*)attrVal,NULL);
                    xmlFree(attrVal);
                  }
                } else
                  xmlFree(attrVal);
              }
            }
          }
          xmlNode = xmlNode->next;
        }
        
        if ( (newCache = IPFAuthCacheCreate(type,size,ttl)) ) {
          if ( adaptive ) {
            IPFAuthCacheSetIsAdaptive(newCache,1,growBy,criticalFrac);
          }
          IPFAuthCacheSetHonorIPPort(newCache, honorIPPort);
        }
      } else {
        xmlFree(attrVal);
      }
    }
  }
  return newCache;
}


#endif /* HAS_LIBXML2 */
