/*
//  pgipfauth - Postgres IPF Authorization Helper
//  IPFAuthCache
//
//  Authorized IP address caching.
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: IPFAuthCache.h 117 2009-03-01 18:32:17Z frey $
*/

#ifndef __IPFAUTHCACHE__
#define __IPFAUTHCACHE__

#include "IPFAuthBase.h"


/*!
  @typedef IPFAuthCacheType
  
  The different types of caches that are implemented.  Type basically equates
  to how the cache is searched.
  
  The default cache type is newest-first.
  
  The stateful cache type does searches rooted at the last-added or last-looked-up
  cache node (whichever was the last operation applied to the cache).  If that
  node has since been purged, a newest-first search is performed.  If that node
  doesn't match the sought-for IP, then the search is bi-directional:  with each
  pass of the search loop two nodes are tested.  The stateful cache type would be
  most useful (hypothetically) in situations where several back-to-back packet
  exchanges are expected intially, with the IP going somewhat silent thereafter.
  Our license servers will possibly see this kind of packet profile for the
  lmgrd and vendor daemons.
  
  Cache type can actually be changed on-the-fly in this implementation.
*/
typedef enum {
  kIPFAuthCacheTypeDefault        = 0x01,
  kIPFAuthCacheTypeOldestFirst    = 0x02,
  kIPFAuthCacheTypeNewestFirst    = 0x03,
  kIPFAuthCacheTypeStateful       = 0x04
} IPFAuthCacheType;


/*!
  @typedef IPFAuthCacheRef
  
  Opaque reference to an IPFAuthCache object.
*/
typedef const struct __IPFAuthCache * IPFAuthCacheRef;


/*!
  @function IPFAuthCacheCreate
  
  Allocate and initialize a new cache with the specified number of
  cache slots and a default time-to-live of ttl for cached addresses.
  
  If type is not one of the values from the IPFAuthCacheType enumeration,
  the default cache type will be chosen.
  
  Returns NULL if the object could not be created.
*/
IPFAuthCacheRef IPFAuthCacheCreate(
    IPFAuthCacheType  type,
    unsigned int      size,
    time_t            ttl
  );


/*!
  @function IPFAuthCacheRetain
  
  Retain a reference to the IPFAuthCache object.
*/
IPFAuthCacheRef IPFAuthCacheRetain(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheRelease
  
  Release a reference to the IPFAuthCache object.  When no more
  references remain, the object is destroyed.
*/
void IPFAuthCacheRelease(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheGetType
  
  Returns the cache type for anAuthCache.
*/
IPFAuthCacheType IPFAuthCacheGetType(
    IPFAuthCacheRef   anAuthCache
  );


/*!
  @function IPFAuthCacheSetType
  
  Sets anAuthCache to be of type anAuthCache.  If type is not one of
  the values from the IPFAuthCacheType enumeration, the default
  cache type will be chosen.
*/
void IPFAuthCacheSetType(
    IPFAuthCacheRef   anAuthCache,
    IPFAuthCacheType  type
  );


/*!
  @function IPFAuthCacheGetSize
  
  Returns the number of cache slots allocated for anAuthCache.
*/
unsigned int IPFAuthCacheGetSize(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheGetTTL
  
  Returns the time-to-live for cached addresses.
*/
time_t IPFAuthCacheGetTTL(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheSetTTL
  
  Modify the time-to-live for cached addresses.  The ttl must be
  greater than or equal to the minimum lifetime -- by default, five
  minutes (decided by the compile-time macro IPFAUTHCACHE_MINIMUM_TTL).
*/
void IPFAuthCacheSetTTL(
    IPFAuthCacheRef anAuthCache,
    time_t          ttl
  );


/*!
  @function IPFAuthCacheIsAdaptive
  
  Returns non-zero if the cache is currently set to function in
  adaptive mode.
*/
int IPFAuthCacheIsAdaptive(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheSetIsAdaptive
  
  If isAdaptive is non-zero, then enable adaptive mode on anAuthCache.
  The cache should increase in size by growBy nodes when the cache is full
  and the fraction of cache misses exceeds atMissFraction.
  
  By default, caches do NOT have adaptive mode enabled.
  
  
*/
void IPFAuthCacheSetIsAdaptive(
    IPFAuthCacheRef anAuthCache,
    int             isAdaptive,
    unsigned int    growBy,
    float           atCriticalPct
  );


/*!
  @function IPFAuthCacheGetHonorIPPort
  
  Returns non-zero if anAuthCache is set to keep and honor the inbound
  IP port.
*/
int IPFAuthCacheGetHonorIPPort(
    IPFAuthCacheRef   anAuthCache
  );


/*!
  @function IPFAuthCacheSetHonorIPPort
  
  Pass a non-zero honorIPPort if anAuthCache should keep and honor the
  inbound IP port.  When this is changed the cache is automatically
  purged.
*/
void IPFAuthCacheSetHonorIPPort(
    IPFAuthCacheRef   anAuthCache,
    int               honorIPPort
  );


/*!
  @function IPFAuthCachePurgeExpired
  
  Expunge any cached IP addresses for which the ttl period has now
  passed.
*/
void IPFAuthCachePurgeExpired(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCachePurgeAll
  
  Expunge all cached IP addresses.
*/
void IPFAuthCachePurgeAll(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheInvalidateIP
  
  Invalidate any cache entries for a specific IP address.
*/
void IPFAuthCacheInvalidateIP(
    IPFAuthCacheRef anAuthCache,
    struct in_addr* ipAddress
  );


/*!
  @function IPFAuthCacheIsAuthorized
  
  Returns non-zero if the specified ipAddress is in the cache; the value
  pointed-to by isAuthorized is set to the cached authorization state.
*/
int IPFAuthCacheIsAuthorized(
    IPFAuthCacheRef       anAuthCache,
    struct in_addr*       ipAddress,
    int                   ipPort,
    IPFAuthDisposition*   isAuthorized
  );


/*!
  @function IPFAuthCacheAuthorize
  
  Add the specified ipAddress to the cache; isAuthorized should be
  a value representing the authorization state of the IP.
*/
void IPFAuthCacheAuthorize(
    IPFAuthCacheRef       anAuthCache,
    struct in_addr*       ipAddress,
    int                   ipPort,
    IPFAuthDisposition    isAuthorized
  );


/*!
  @function IPFAuthCacheDebug
  
  Write a debugging-esque summary of the cache to stderr.
*/
void IPFAuthCacheDebug(
    IPFAuthCacheRef anAuthCache
  );


/*!
  @function IPFAuthCacheInfo
  
  Write a description (less verbose than the debug output) of the cache to stdout.
*/
void
IPFAuthCacheInfo(
    IPFAuthCacheRef anAuthCache
  );


#ifdef HAS_LIBXML2
#include <libxml/tree.h>


/*!
  @function IPFAuthCacheCreateWithXMLNode
  
  Create a new cache object using the properties specified in a "cache"
  element at the same document level as xmlNode.  See the included
  pgipfauth-conf_1.0.dtd for the cache element format.
*/
IPFAuthCacheRef IPFAuthCacheCreateWithXMLNode(
    xmlNodePtr      xmlNode
  );


#endif /* HAS_LIBXML2 */


#endif /* __IPFAUTHCACHE_ */
