/*
//  pgipfauth - Postgres IPF Authorization Helper
//  IPFDatabase
//
//  A pseudo-class that handles all interaction with the Postgres
//  database that backs this authorization mechanism.
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: IPFDatabase.c 117 2009-03-01 18:32:17Z frey $
*/

#include <stdio.h>
#include <ctype.h>

#include "IPFDatabase.h"

#define PGSQL_AF_INET   (AF_INET + 0)
#define PGSQL_AF_INET6  (AF_INET + 1)

typedef struct {
  unsigned char      family;
  unsigned char      bits;
  unsigned char      is_cidr;
  unsigned char      addr_length;
  unsigned char      addr[4];
} IPFDatabasePGINET;

void IPFDatabasePGINETSet(
  IPFDatabasePGINET*    pgInet,
  struct in_addr*       ipaddr
)
{
  pgInet->family = PGSQL_AF_INET;
  pgInet->bits = 32;
  pgInet->is_cidr = 0;
  pgInet->addr_length = 4;
  memcpy(&pgInet->addr[0], ipaddr, 4);
}

//

/*!
  @constant IPFDatabaseDefaultSchema
  
  By default, we'll assume the tables are defined in the default schema.
*/
const char* IPFDatabaseDefaultSchema = "<<<NO SCHEMA>>>";


/*!
  @constant IPFDatabaseLookupPrepStmt
  
  The name of our prepared statement for IP lookup.
*/
const char* IPFDatabaseLookupPrepStmt = "IPF_lookupIPAddress";


/*!
  @enum IPFDatabase flags
  
  Internal flags that alter the functioning of an IPF database object.
*/
enum {
  kIPFDatabaseHasHostGroups    = 1 << 0,
  kIPFDatabaseUsesHostGroup    = 1 << 1,
  kIPFDatabaseNeedsReset       = 1 << 16
};


/*!
  @typedef IPFDatabaseLookupCallback
  
  We adapt to different kinds of lookup by using the function pointer of
  the option-appropriate method.  The IPFDatabaseLookupCallback points to such
  a function.
*/
typedef IPFAuthDisposition (*IPFDatabaseLookupCallback)(struct __IPFDatabase* aDatabase,struct in_addr* ipAddress,int ipPort);


/*!
  @typedef IPFDatabase
  
  A database pseudo-object.
*/
typedef struct __IPFDatabase {
  unsigned int                      refCount;
  unsigned int                      flags;
  
  const char*                       connStrCopy;
  const char*                       schema;
  PGconn*                           dbHandle;
  
  int                               hgId,hgIdSwapped;
  const char*                       hostGroup;
  
  IPFDatabaseLookupCallback         lookupProc;
  
  IPFDatabaseResetListenerCallback  resetListenerProc;
  void*                             resetListenerContext;
} IPFDatabase;

int __IPFDatabase_ResetConnection(
    IPFDatabase*    aDatabase
  );

/*!
  @function __IPFDatabase_LookupSansHostGroup
  
  Lookup ipAddress for authorization in a database schema that has
  no hostGroup table (and thus, no sense of host groups).
  
  Returns non-zero if ipAddress is authorized, zero if not.
  
  If the query yields a "bad response" or "fatal error" then
  we'll attempt to re-connect to the database and reset the
  database object to reflect any changes to the database.
*/
IPFAuthDisposition
__IPFDatabase_LookupSansHostGroup(
  IPFDatabase*    aDatabase,
  struct in_addr* ipAddress,
  int             ipPort
)
{
  PGresult*           lookupResult = NULL;
  IPFDatabasePGINET   inetForm;
  IPFAuthDisposition  rc = kIPFAuthDisposition_Block;
  char*               params[2];
  int                 lengths[2];
  int                 formats[2];
  
  /* Need to byte-swap the port: */
  ipPort = htonl(ipPort);
  
  params[0] = (char*)&inetForm;       lengths[0] = sizeof(IPFDatabasePGINET);     formats[0] = 1;
  params[1] = (char*)&ipPort;         lengths[1] = sizeof(int);                   formats[1] = 1;
  
  IPFDatabasePGINETSet(&inetForm, ipAddress);
  lookupResult = PQexecPrepared(
                      aDatabase->dbHandle,
                      IPFDatabaseLookupPrepStmt,
                      2,
                      (const char* const*)params,
                      (const int*)lengths,
                      (const int*)formats,
                      1
                    );
  if ( PQstatus(aDatabase->dbHandle) != CONNECTION_OK ) {
    /* Try to reset the connection and then re-authorize: */
    if ( __IPFDatabase_ResetConnection(aDatabase) )
      rc = IPFDatabaseAuthorizeIP((IPFDatabaseRef)aDatabase, ipAddress, ipPort);
    else
      rc = kIPFAuthDisposition_Unknown;
    if ( lookupResult ) PQclear(lookupResult);
  } else if ( lookupResult ) {
    switch ( PQresultStatus(lookupResult) ) {
    
      case PGRES_TUPLES_OK: {
        if ( PQntuples(lookupResult) > 0 ) {
          //  Boolean types are represented as a single-byte:
          char*       singleByte = (char*)PQgetvalue(lookupResult,0,0);
          
          if ( *singleByte )
            rc = kIPFAuthDisposition_Pass;
        }
        break;
      }
      
    }
    PQclear(lookupResult);
  }
  
  return rc;
}


/*!
  @function __IPFDatabase_LookupWithHostGroup
  
  Lookup ipAddress for authorization with respect to a particular
  host group.  If the database object was created with NO host
  group and we're using this routine, that implies that the database
  is host-group-capable, and thus we should use a "default" host
  group ID -- which is zero.
  
  Returns non-zero if ipAddress is authorized, zero if not.
  
  If the query yields a "bad response" or "fatal error" then
  we'll attempt to re-connect to the database and reset the
  database object to reflect any changes to the database.
*/
IPFAuthDisposition
__IPFDatabase_LookupWithHostGroup(
  IPFDatabase*    aDatabase,
  struct in_addr* ipAddress,
  int             ipPort
)
{
  PGresult*           lookupResult = NULL;
  IPFDatabasePGINET   inetForm;
  IPFAuthDisposition  rc = kIPFAuthDisposition_Block;
  char*               params[3];
  int                 lengths[3];
  int                 formats[3];
  int                 hostGroup = ( aDatabase->flags & kIPFDatabaseHasHostGroups ? aDatabase->hgIdSwapped : 0 );
  
  /* Need to byte-swap the port: */
  ipPort = htonl(ipPort);
  
  params[0] = (char*)&hostGroup;      lengths[0] = sizeof(int);                   formats[0] = 1;
  params[1] = (char*)&inetForm;       lengths[1] = sizeof(IPFDatabasePGINET);     formats[1] = 1;
  params[2] = (char*)&ipPort;         lengths[2] = sizeof(int);                   formats[2] = 1;
  
  lookupResult = PQexecPrepared(
                      aDatabase->dbHandle,
                      IPFDatabaseLookupPrepStmt,
                      3,
                      (const char* const*)params,
                      (const int*)lengths,
                      (const int*)formats,
                      1
                    );
  if ( PQstatus(aDatabase->dbHandle) != CONNECTION_OK ) {
    /* Try to reset the connection and then re-authorize: */
    if ( __IPFDatabase_ResetConnection(aDatabase) )
      rc = IPFDatabaseAuthorizeIP((IPFDatabaseRef)aDatabase, ipAddress, ipPort);
    else
      rc = kIPFAuthDisposition_Unknown;
    if ( lookupResult ) PQclear(lookupResult);
  } else if ( lookupResult ) {
    switch ( PQresultStatus(lookupResult) ) {
      
      case PGRES_TUPLES_OK: {
        if ( PQntuples(lookupResult) > 0 ) {
          //  Boolean types are represented as a single-byte:
          char*       singleByte = (char*)PQgetvalue(lookupResult,0,0);
          
          if ( *singleByte )
            rc = kIPFAuthDisposition_Pass;
        }
        break;
      }
      
    }
    PQclear(lookupResult);
  }
  return rc;
}


/*!
  @function __IPFDatabase_SetSearchPath
  
  Set the schema search path on the database to avoid having to specify the schema in
  our queries.
*/
int
__IPFDatabase_SetSearchPath(
  IPFDatabase*      aDatabase
)
{
  static const char*    formatStr = "SET search_path TO \"%s\",\"$user\",public;";
  
  if ( strcmp(aDatabase->schema,IPFDatabaseDefaultSchema) != 0 ) {
    int                 rc = 0;
    size_t              queryLen = strlen(formatStr) + strlen(aDatabase->schema) - 2 + 1;
    char*               query = malloc(queryLen);
    
    if ( query ) {
      PGresult*         spResult;
      
      snprintf(query,queryLen,formatStr,aDatabase->schema);
      if ( spResult = PQexec(aDatabase->dbHandle,query) ) {
        if ( PQresultStatus(spResult) == PGRES_COMMAND_OK ) {
          rc = 1;
        } else {
          fprintf(stderr,"ERROR:  unable to set schema search path in database\n");
        }
        PQclear(spResult);
      } else {
        fprintf(stderr,"ERROR:  unable to set schema search path in database\n");
      }
      free(query);
    }
    return rc;
  }
  return 1;
}


/*!
  @function __IPFDatabase_GetHostGroupID
  
  Lookup the textually-named host group provided when aDatabase was
  created and return the integral identifier assigned to that group.
*/
int
__IPFDatabase_GetHostGroupID(
  IPFDatabase*      aDatabase
)
{
  int               hgId = -1;
  PGresult*         hgQueryResult = PQexecParams(
                                        aDatabase->dbHandle,
                                        "SELECT hgId FROM hostGroup WHERE name = $1",
                                        1,
                                        NULL,
                                        &aDatabase->hostGroup,
                                        NULL,
                                        NULL,
                                        1
                                      );
  
  if ( hgQueryResult ) {
    if ( (PQresultStatus(hgQueryResult) == PGRES_TUPLES_OK) && (PQntuples(hgQueryResult) > 0) ) {
      memcpy(&hgId,PQgetvalue(hgQueryResult,0,0),sizeof(int));
      //  Byte swap:
      hgId = ntohl(hgId);
    }
    PQclear(hgQueryResult);
  }
  return hgId;
}


/*!
  @function __IPFDatabase_CheckForHostGroup
  
  Check the pg_tables system table for the "hostGroup" table; if it's there
  we'll assume that we're setup to do host groups.
  
  If an actual hostGroup was provided, we'll lookup its numeric ID and setup
  the object's flags and lookup callback.
*/
int
__IPFDatabase_CheckForHostGroup(
  IPFDatabase*      aDatabase
)
{
  PGresult*         hgQueryResult = NULL;
  
  if ( strcmp(aDatabase->schema,IPFDatabaseDefaultSchema) == 0 ) {
    hgQueryResult = PQexec(
                          aDatabase->dbHandle,
                          "SELECT * FROM pg_tables WHERE schemaName = '$user' OR schemaName = 'public' AND tableName ILIKE 'hostGroup'"
                        );
  } else {
    hgQueryResult = PQexecParams(
                          aDatabase->dbHandle,
                          "SELECT * FROM pg_tables WHERE schemaName ILIKE $1 OR schemaName = '$user' OR schemaName = 'public' AND tableName ILIKE 'hostGroup'",
                          1,
                          NULL,
                          &aDatabase->schema,
                          NULL,
                          NULL,
                          0
                        );
  }
  
  aDatabase->flags &= ~(kIPFDatabaseHasHostGroups | kIPFDatabaseUsesHostGroup);
  
  if ( hgQueryResult ) {
    if ( (PQresultStatus(hgQueryResult) == PGRES_TUPLES_OK) && (PQntuples(hgQueryResult) > 0) ) {
      PQclear(hgQueryResult);
      
      aDatabase->flags |= kIPFDatabaseHasHostGroups;
      
      /* Now, resolve the host group to a numerical identifier: */
      if ( (aDatabase->hgId = __IPFDatabase_GetHostGroupID(aDatabase)) >= 0 ) {
        aDatabase->hgIdSwapped = htonl(aDatabase->hgId);
        aDatabase->flags |= kIPFDatabaseUsesHostGroup;
      } else {
        fprintf(stderr,"WARNING:  unknown host group: %s\n",aDatabase->hostGroup);
      }
    } else
      PQclear(hgQueryResult);
  
    if ( aDatabase->flags & kIPFDatabaseHasHostGroups )
      aDatabase->lookupProc = __IPFDatabase_LookupWithHostGroup;
    else
      aDatabase->lookupProc = __IPFDatabase_LookupSansHostGroup;
    
    return 1;
  }
  return 0;
}


/*!
  @function __IPFDatabase_PrepareQuery
  
  Rather than repeatedly issuing the same IP lookup query, we use a prepared
  query -- as prepped by this function.
*/
int
__IPFDatabase_PrepareQuery(
  IPFDatabase*    aDatabase
)
{
  PGresult*       prepResult = NULL;
  int             rc = 0;
  
  if ( aDatabase->flags & kIPFDatabaseHasHostGroups ) {
    prepResult = PQprepare(
                    aDatabase->dbHandle,
                    IPFDatabaseLookupPrepStmt,
                    "SELECT pgipfAuthorizeWithHostGroup($1,$2,$3)",
                    0,
                    NULL
                  );
  } else {
    prepResult = PQprepare(
                    aDatabase->dbHandle,
                    IPFDatabaseLookupPrepStmt,
                    "SELECT pgipfAuthorize($1,$2)",
                    0,
                    NULL
                  );
  }
  
  if ( prepResult ) {
    if ( PQresultStatus(prepResult) == PGRES_COMMAND_OK ) {
      rc = 1;
    } else {
      fprintf(stderr,"ERROR:  unable to prepare our authorization query in database\n");
    }
    PQclear(prepResult);
  } else {
    fprintf(stderr,"ERROR:  unable to prepare our authorization query in database\n");
  }
  return rc;
}


/*!
  @function __IPFDatabase_ResetConnection
  
  Reset the database connection; if the reset yields a good connection,
  again set the schema search path, check for host group stuff, and
  prep the authorization query.
  
  Returns non-zero if the reset was totally successful.
*/
int
__IPFDatabase_ResetConnection(
  IPFDatabase*    aDatabase
)
{
  int             rc = 0;
  
  aDatabase->flags |= kIPFDatabaseNeedsReset;
  
  PQreset(aDatabase->dbHandle);
  if ( PQstatus(aDatabase->dbHandle) == CONNECTION_OK ) {
    if ( __IPFDatabase_SetSearchPath(aDatabase) ) {
      if ( __IPFDatabase_CheckForHostGroup(aDatabase) ) {
        if ( __IPFDatabase_PrepareQuery(aDatabase) ) {
          aDatabase->flags &= ~kIPFDatabaseNeedsReset;
          rc = 1;
        }
      }
    }
  }
  if ( aDatabase->resetListenerProc )
    aDatabase->resetListenerProc((IPFDatabaseRef)aDatabase,aDatabase->resetListenerContext);
  
  return rc;
}


/* **** */


IPFDatabaseRef
IPFDatabaseCreate(
  const char*     connStr,
  const char*     schema
)
{
  return IPFDatabaseCreateWithHostGroup(connStr,schema,NULL);
}


IPFDatabaseRef
IPFDatabaseCreateWithHostGroup(
  const char*     connStr,
  const char*     schema,
  const char*     hostGroup
)
{
  IPFDatabase*    newDB = NULL;
  PGconn*         dbHandle = PQconnectdb(connStr);
  
  if ( ! schema || (strlen(schema) == 0) )
    schema = IPFDatabaseDefaultSchema;
  
  if ( dbHandle ) {
    if ( PQstatus(dbHandle) == CONNECTION_OK ) {
      size_t      extraBytes = strlen(connStr) + strlen(schema) + 2;
      
      if ( hostGroup ) {
        extraBytes += strlen(hostGroup) + 1;
      }
      
      if ( newDB = calloc(1,sizeof(IPFDatabase) + extraBytes) ) {
        newDB->refCount = 1;
        
        newDB->connStrCopy = ((char*)newDB) + sizeof(IPFDatabase);
        strcpy((char*)newDB->connStrCopy,connStr);
        
        newDB->dbHandle = dbHandle;
        
        newDB->schema = newDB->connStrCopy + strlen(connStr) + 1;
        strcpy((char*)newDB->schema,schema);
        
        if ( hostGroup ) {
          newDB->hostGroup = newDB->schema + strlen(schema) + 1;
          strcpy((char*)newDB->hostGroup,hostGroup);
        }
        
        /* Set the search path for our session: */
        if ( __IPFDatabase_SetSearchPath(newDB) ) {
          if ( __IPFDatabase_CheckForHostGroup(newDB) ) {
            //  Try preparing our query now; if that fails, we're S.O.L.
            if ( ! __IPFDatabase_PrepareQuery(newDB) ) {
              IPFDatabaseRelease((IPFDatabaseRef)newDB);
              newDB = NULL;
            }
          } else {
            IPFDatabaseRelease((IPFDatabaseRef)newDB);
            newDB = NULL;
          }
        } else {
          IPFDatabaseRelease((IPFDatabaseRef)newDB);
          newDB = NULL;
        }
      } else {
        PQfinish(dbHandle);
      }
    } else {
      PQfinish(dbHandle);
    }
  }
  return newDB;
}
  

IPFDatabaseRef
IPFDatabaseRetain(
  IPFDatabaseRef  aDatabase
)
{
  ((IPFDatabase*)aDatabase)->refCount++;
}


void
IPFDatabaseRelease(
  IPFDatabaseRef  aDatabase
)
{
  if ( --((IPFDatabase*)aDatabase)->refCount == 0 ) {
    PQfinish(aDatabase->dbHandle);
    free((IPFDatabase*)aDatabase);
  }
}


IPFDatabaseResetListenerCallback
IPFDatabaseGetResetListener(
  IPFDatabaseRef  aDatabase
)
{
  return aDatabase->resetListenerProc;
}
  

void
IPFDatabaseSetResetListener(
  IPFDatabaseRef  aDatabase,
  IPFDatabaseResetListenerCallback listener,
  void*           context
)
{
  ((IPFDatabase*)aDatabase)->resetListenerProc = listener;
  ((IPFDatabase*)aDatabase)->resetListenerContext = context;
}


const char*
IPFDatabaseGetConnStr(
  IPFDatabaseRef  aDatabase
)
{
  return aDatabase->connStrCopy;
}

  
PGconn*
IPFDatabaseGetPGHandle(
  IPFDatabaseRef  aDatabase
)
{
  return aDatabase->dbHandle;
}


IPFAuthDisposition
IPFDatabaseAuthorizeIP(
  IPFDatabaseRef  aDatabase,
  struct in_addr* anIPAddress,
  int             anIPPort
)
{
  if ( aDatabase->flags & kIPFDatabaseNeedsReset ) {
    if ( ! __IPFDatabase_ResetConnection((IPFDatabase*)aDatabase) ) {
      return kIPFAuthDisposition_Unknown;
    }
  }
  return (aDatabase->lookupProc)((IPFDatabase*)aDatabase, anIPAddress, anIPPort);
}


void
IPFDatabaseDebug(
  IPFDatabaseRef  aDatabase
)
{
  fprintf(stderr,
        "IPFDatabase@%p [%u] {\n"
        "  connection string = %s\n"
        "  schema            = %s\n"
        "  flags             = 0x%04X\n",
        aDatabase,aDatabase->refCount,
        aDatabase->connStrCopy,
        aDatabase->schema,
        aDatabase->flags
      );
  if ( aDatabase->hgId > 0 ) {
    fprintf(stderr,"  host group        = id: %d; name: %s\n",aDatabase->hgId,aDatabase->hostGroup);
  }
  fprintf(stderr,"}\n");
}


void
IPFDatabaseInfo(
  IPFDatabaseRef  aDatabase
)
{
  printf("  database summary {\n"
         "    connection string: \"%s\"\n"
         "    schema: %s\n"
         "    host groups {\n"
         "      available: %s\n",
         aDatabase->connStrCopy,
         aDatabase->schema,
         ( aDatabase->flags & kIPFDatabaseHasHostGroups ? "yes" : "no" )
      );
  if ( aDatabase->flags & kIPFDatabaseUsesHostGroup ) {
    printf("      using: %s (id = %d)\n",
           aDatabase->hostGroup,aDatabase->hgId
        );
  }
  printf("    }\n"
         "    needs-reset: %s\n"
         "  }\n",
      ( aDatabase->flags & kIPFDatabaseNeedsReset ? "yes" : "no" )
    );
}

#ifdef HAS_LIBXML2
#include <libxml/parser.h>

const xmlChar* IPFDatabase_xml_database = BAD_CAST("database");
const xmlChar* IPFDatabase_xml_database_sslmode = BAD_CAST("sslmode");
const xmlChar* IPFDatabase_xml_database_sslmode_disable = BAD_CAST("disable");
const xmlChar* IPFDatabase_xml_database_sslmode_allow = BAD_CAST("allow");
const xmlChar* IPFDatabase_xml_database_sslmode_prefer = BAD_CAST("prefer");
const xmlChar* IPFDatabase_xml_database_sslmode_require = BAD_CAST("require");

const xmlChar* IPFDatabase_xml_host = BAD_CAST("host");
const xmlChar* IPFDatabase_xml_hostaddr = BAD_CAST("hostaddr");
const xmlChar* IPFDatabase_xml_port = BAD_CAST("port");
const xmlChar* IPFDatabase_xml_dbname = BAD_CAST("dbname");
const xmlChar* IPFDatabase_xml_user = BAD_CAST("user");
const xmlChar* IPFDatabase_xml_password = BAD_CAST("password");
const xmlChar* IPFDatabase_xml_password_type = BAD_CAST("type");
const xmlChar* IPFDatabase_xml_password_type_inline = BAD_CAST("inline");
const xmlChar* IPFDatabase_xml_password_type_external = BAD_CAST("external");
const xmlChar* IPFDatabase_xml_schema = BAD_CAST("schema");
const xmlChar* IPFDatabase_xml_hostGroup = BAD_CAST("host-group");

IPFDatabaseRef
IPFDatabaseCreateWithXMLNode(
  xmlNodePtr      xmlNode
)
{
  IPFDatabaseRef  newDB = NULL;
  xmlNodePtr      altNode;
  
  /* Look for our node: */
  while ( xmlNode ) {
    if ( xmlNode->type == XML_ELEMENT_NODE ) {
      if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_database) == 0 ) {
        break;
      }
    }
    xmlNode = xmlNode->next;
  }
  
  if ( xmlNode ) {
    xmlChar     *host = NULL,*hostaddr = NULL,*dbname = NULL,*user = NULL,*password = NULL,*port = NULL,
                *schema = NULL,*hostGroup = NULL;
    xmlChar*    sslmode = xmlGetProp(xmlNode,IPFDatabase_xml_database_sslmode);
    
    /* Validate the SSL mode: */
    if ( sslmode ) {
      if ( ! ( (xmlStrcasecmp(sslmode,IPFDatabase_xml_database_sslmode_disable) == 0) ||
               (xmlStrcasecmp(sslmode,IPFDatabase_xml_database_sslmode_allow) == 0) ||
               (xmlStrcasecmp(sslmode,IPFDatabase_xml_database_sslmode_prefer) == 0) ||
               (xmlStrcasecmp(sslmode,IPFDatabase_xml_database_sslmode_require) == 0) )
         )
      {
        fprintf(stderr,"WARNING:  invalid sslmode: %s\n",sslmode);
        xmlFree(sslmode);
        sslmode = NULL;
      }
    }
    
    /* Walk the child elements and setup our properties: */
    xmlNode = xmlNode->children;
    while ( xmlNode ) {
      if ( xmlNode->type == XML_ELEMENT_NODE ) {
        /* schema */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_schema) == 0 ) {
          schema = xmlNodeGetContent(xmlNode);
        } else
        
        /* hostGroup */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_hostGroup) == 0 ) {
          hostGroup = xmlNodeGetContent(xmlNode);
        } else
        
        /* host */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_host) == 0 ) {
          host = xmlNodeGetContent(xmlNode);
        } else
        
        /* hostaddr */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_hostaddr) == 0 ) {
          hostaddr = xmlNodeGetContent(xmlNode);
        } else
        
        /* port */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_port) == 0 ) {
          port = xmlNodeGetContent(xmlNode);
        } else
        
        /* dbname */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_dbname) == 0 ) {
          dbname = xmlNodeGetContent(xmlNode);
        } else
        
        /* user */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_user) == 0 ) {
          user = xmlNodeGetContent(xmlNode);
        } else
        
        /* password */
        if ( xmlStrcasecmp(xmlNode->name,IPFDatabase_xml_password) == 0 ) {
          xmlChar*      passwordType = xmlGetProp(xmlNode,IPFDatabase_xml_password_type);
          
          password = xmlNodeGetContent(xmlNode);
          if ( passwordType ) {
            if ( xmlStrcasecmp(passwordType,IPFDatabase_xml_password_type_external) == 0 ) {
              /* Attempt to open a file with password as the path: */
              FILE*     pwdFH = fopen((const char*)password,"r");
              
              if ( pwdFH ) {
                char*   tmpPassword = NULL;
                off_t   start = ftello(pwdFH);
                off_t   end;
                
                fseeko(pwdFH,0,SEEK_END);
                end = ftello(pwdFH);
                fseeko(pwdFH,0,SEEK_SET);
                
                /* Now we know the file length, so allocate space for a full string
                   and read it in -- stripping whitespace and control characters
                   in the process */
                end = end - start + 1;
                if ( tmpPassword = calloc(1,end + 1) ) {
                  char*   s = tmpPassword;
                  int     foundText = 0;
                  
                  while ( end-- ) {
                    char  c = fgetc(pwdFH);
                    if ( isascii(c) && ! ( isspace(c) || iscntrl(c) ) ) {
                      foundText = 1;
                      *s = c;
                      s++;
                    } else if ( foundText ) {
                      break;
                    }
                  }
                  if ( strlen(tmpPassword) == 0 ) {
                    fprintf(stderr,"WARNING:  invalid password (net length of zero) in password file: %s\n",password);
                    xmlFree(password);
                    password = NULL;
                  } else {
                    xmlFree(password);
                    password = xmlStrdup((xmlChar*)tmpPassword);
                  }
                  free(tmpPassword);
                }
                fclose(pwdFH);
              } else {
                fprintf(stderr,"WARNING:  could not read password file: %s\n",password);
                xmlFree(password);
                password = NULL;
              }
            } else
            if ( xmlStrcasecmp(passwordType,IPFDatabase_xml_password_type_inline) != 0 ) {
              fprintf(stderr,"WARNING:  invalid password type attribute: %s\n",passwordType);
            }
            xmlFree(passwordType);
          }
        }
      }
      xmlNode = xmlNode->next;
    }
    
    /* We AT LEAST need a dbname: */
    if ( dbname ) {
      char*     connStr = IPFDatabaseBuildConnStr(
                              (char*)host,
                              (char*)hostaddr,
                              (char*)port,
                              (char*)dbname,
                              (char*)user,
                              (char*)password,
                              (char*)sslmode
                            );
      
      if ( connStr ) {
        newDB = IPFDatabaseCreateWithHostGroup(connStr,(char*)schema,(char*)hostGroup);
        free(connStr);
      }
      xmlFree(dbname);
    } else {
      fprintf(stderr,"ERROR:  no database name provided\n");
    }
    if ( host ) xmlFree(host);
    if ( hostaddr ) xmlFree(hostaddr);
    if ( user ) xmlFree(user);
    if ( password ) xmlFree(password);
    if ( schema ) xmlFree(schema);
    if ( hostGroup ) xmlFree(hostGroup);
  }
  return newDB;
}


#endif /* HAS_LIBXML2 */


char*
IPFDatabaseBuildConnStr(
  char*     host,
  char*     hostaddr,
  char*     port,
  char*     dbname,
  char*     user,
  char*     password,
  char*     sslmode
)
{
  static char*    partFmt = "%s='%s' %n";
  static size_t   partBaseLen = 4;
  
  char*     result = NULL;
  size_t    resultLen = 1;
  
  if ( host )
    resultLen += partBaseLen + strlen("host") + strlen(host);
  if ( hostaddr )
    resultLen += partBaseLen + strlen("hostaddr") + strlen(hostaddr);
  if ( port )
    resultLen += partBaseLen + strlen("port") + strlen(port);
  if ( dbname )
    resultLen += partBaseLen + strlen("dbname") + strlen(dbname);
  if ( user )
    resultLen += partBaseLen + strlen("user") + strlen(user);
  if ( password )
    resultLen += partBaseLen + strlen("password") + strlen(password);
  if ( sslmode )
    resultLen += partBaseLen + strlen("sslmode") + strlen(sslmode);
  
  if ( result = calloc(1,resultLen) ) {
    int       n;
    char*     base = result;
    
    if ( host ) {
      snprintf(base,resultLen,partFmt,"host",host,&n);
      base += n; resultLen -= n;
    }
    if ( hostaddr ) {
      snprintf(base,resultLen,partFmt,"hostaddr",hostaddr,&n);
      base += n; resultLen -= n;
    }
    if ( port ) {
      snprintf(base,resultLen,partFmt,"port",port,&n);
      base += n; resultLen -= n;
    }
    if ( dbname ) {
      snprintf(base,resultLen,partFmt,"dbname",dbname,&n);
      base += n; resultLen -= n;
    }
    if ( user ) {
      snprintf(base,resultLen,partFmt,"user",user,&n);
      base += n; resultLen -= n;
    }
    if ( password ) {
      snprintf(base,resultLen,partFmt,"password",password,&n);
      base += n; resultLen -= n;
    }
    if ( sslmode ) {
      snprintf(base,resultLen,partFmt,"sslmode",sslmode,&n);
      base += n; resultLen -= n;
    }
  }
  
  return result;
}
