/*
//  pgipfauth - Postgres IPF Authorization Helper
//  IPFDatabase
//
//  A pseudo-class that handles all interaction with the Postgres
//  database that backs this authorization mechanism.
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: IPFDatabase.h 117 2009-03-01 18:32:17Z frey $
*/

#ifndef __IPFDATABASE__
#define __IPFDATABASE__


#include "libpq-fe.h"
#include "IPFAuthBase.h"


/*!
  @constant IPFDatabaseDefaultSchema
  
  The default database schema that contains the authorization table.  Guess what,
  it's just "public" *grin*
*/
const char* IPFDatabaseDefaultSchema;


/*!
  @typedef IPFDatabaseRef
  
  Opaque reference to an IPFDatabase object.
*/
typedef const struct __IPFDatabase * IPFDatabaseRef;


/*!
  @typedef IPFDatabaseResetListenerCallback
  
  Pointer to a function that an IPFDatabase instance will call when it needs to
  reconnect to its Postgres database.  The ipfauth daemon uses this mechanism to
  automatically flush the cache when a reconnect event happens.
  
  The function should take two arguments:  a reference to the IPFDatabase object
  that just tried a reconnection and a user-defined pointer value that was
  provided to the IPFDatabase when the listener procedure was registered.
*/
typedef void (*IPFDatabaseResetListenerCallback)(IPFDatabaseRef aDatabase,void* context);


/*!
  @function IPFDatabaseCreate
  
  Allocate and initialize a database object using a Postgres "connection string"
  and a database schema name.
  
  A convenience, this function merely calls IPFDatabaseCreateWithHostGroup() with
  a hostGroup of NULL.
*/
IPFDatabaseRef IPFDatabaseCreate(
    const char*     connStr,
    const char*     schema
  );


/*!
  @function IPFDatabaseCreateWithHostGroup
  
  Allocate and initialize a database object using a Postgres "connection string",
  a database schema name, and a hostGroup name.
  
  The schema may be NULL, which is interpreted as being equivalent to passing
  IPFDatabaseDefaultSchema.
  
  The hostGroup may be NULL, which is interpreted as meaning the object should
  not restrict itself to a particular host group when authorizing an IP
  address.
*/
IPFDatabaseRef IPFDatabaseCreateWithHostGroup(
    const char*     connStr,
    const char*     schema,
    const char*     hostGroup
  );


/*!
  @function IPFDatabaseRetain
  
  Retain a reference to the IPFDatabase object.
*/
IPFDatabaseRef IPFDatabaseRetain(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseRelease
  
  Release a reference to the IPFDatabase object.  When no more
  references remain, the database connection is closed and the object
  is destroyed.
*/
void IPFDatabaseRelease(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseGetResetListener
  
  Returns the currently-registered reset-event listener function
  pointer.
*/
IPFDatabaseResetListenerCallback IPFDatabaseGetResetListener(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseSetResetListener
  
  Register a reset-event listener function (complete with a user-defined
  pointer value, context) that will be called when the database connection
  must be reset.
*/
void IPFDatabaseSetResetListener(
    IPFDatabaseRef  aDatabase,
    IPFDatabaseResetListenerCallback listener,
    void*           context
  );


/*!
  @function IPFDatabaseGetConnStr
  
  Each database object makes its own copy of the original connection string
  that was passed to its *Create() function.  This function returns a pointer
  to that string.
*/
const char* IPFDatabaseGetConnStr(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseGetPGHandle
  
  Returns the Postgres database handle which is being used by this
  IPFDatabase object.
*/
PGconn* IPFDatabaseGetPGHandle(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseAuthorizeIP
  
  Returns the authorization state of anIPAddress reflected in the current
  database state.
  
  If the connection to the database has failed and cannot be re-established,
  return kIPFAuthDisposition_Unknown.
*/
IPFAuthDisposition IPFDatabaseAuthorizeIP(
    IPFDatabaseRef  aDatabase,
    struct in_addr* anIPAddress,
    int             anIPPort
  );

/*!
  @function IPFDatabaseDebug
  
  Write a debugging-esque summary of the database to stderr.
*/
void IPFDatabaseDebug(
    IPFDatabaseRef  aDatabase
  );


/*!
  @function IPFDatabaseInfo
  
  Write a description (less verbose than the debug output) of the cache to stdout.
*/
void IPFDatabaseInfo(
    IPFDatabaseRef  aDatabase
  );


#ifdef HAS_LIBXML2
#include <libxml/tree.h>


/*!
  @function IPFDatabaseCreateWithXMLNode
  
  Create a new database object using the properties specified in a "database"
  element at the same document level as xmlNode.  See the included
  pgipfauth-conf_1.0.dtd for the cache element format.
*/
IPFDatabaseRef IPFDatabaseCreateWithXMLNode(
    xmlNodePtr      xmlNode
  );


#endif /* HAS_LIBXML2 */


/*!
  @function IPFDatabaseBuildConnStr
  
  Returns a newly-allocated character string that contains the Postgres
  connection string associated with the passed arguments.
  
  Any NULL parameters will not appear in the string; any parameters that
  are empty strings will appear as empty strings.
  
  The caller is responsible for free'ing the string (eventually).
*/
char* IPFDatabaseBuildConnStr(
    char*     host,
    char*     hostaddr,
    char*     port,
    char*     dbname,
    char*     user,
    char*     password,
    char*     sslmode
  );

#endif /* __IPFDATABASE__ */
