/*
//  pgipfauth - Postgres IPF Authorization Helper
//  ipfcacheinvalidate.c
//
//  Postgres external trigger function which reacts to changes to the
//  authorization table by writing to the cache-invalidation FIFO
//  which pgipfauth is watching. 
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: ipfcacheinvalidate.c 119 2009-03-04 16:51:03Z frey $
*/

#include "postgres.h"
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */


#ifndef PGIPFAUTH_INVALIDATE_FIFO
#define PGIPFAUTH_INVALIDATE_FIFO "/usr/local/pgipfauth/current/etc/cache-invalidate"
#endif

#ifndef PGIPFAUTH_REMOTEIP_FIELD
#define PGIPFAUTH_REMOTEIP_FIELD "remoteaddress"
#endif

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

extern Datum ipfcacheinvalidate(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(ipfcacheinvalidate);

Datum
ipfcacheinvalidate(PG_FUNCTION_ARGS)
{
  TriggerData   *triggerData = (TriggerData *) fcinfo->context;
  TupleDesc     tupleDesc;
  HeapTuple     resultTuple;
  int           fnum;

  /* make sure it's called as a trigger */
  if (!CALLED_AS_TRIGGER(fcinfo))
    elog(ERROR, "ipfcacheinvalidate: not called by trigger manager");

  tupleDesc = triggerData->tg_relation->rd_att;
  resultTuple = triggerData->tg_trigtuple;
  
  /* Find the appropriate field: */
  fnum = SPI_fnumber(tupleDesc, PGIPFAUTH_REMOTEIP_FIELD);
  if ( fnum >= 0 ) {
    char*       ipAddress = SPI_getvalue(resultTuple, tupleDesc, fnum);
    
    if ( ipAddress && strlen(ipAddress) ) {
      FILE*     cacheFIFO = fopen(PGIPFAUTH_INVALIDATE_FIFO, "w");
      
      if ( cacheFIFO ) {
        /* On UPDATE, let's do BOTH ipAddresses: */
        if ( TRIGGER_FIRED_BY_UPDATE(triggerData->tg_event) ) {
          char* newIPAddress = SPI_getvalue(triggerData->tg_newtuple, tupleDesc, fnum);
          
          fprintf(cacheFIFO, "%s %s\n", ipAddress, (newIPAddress && strlen(newIPAddress) ? newIPAddress : "") );
        } else {
          fprintf(cacheFIFO, "%s\n", ipAddress);
        }
        fclose(cacheFIFO);
      } else {
        elog(INFO, "ipfcacheinvalidate: unable to write to cache FIFO");
      }
    }
  }
  
  /* All done: */
  return PointerGetDatum(resultTuple);
}

/*

CREATE FUNCTION ipfcacheinvalidate() RETURNS TRIGGER AS '/usr/local/pgipfauth/0.1/src/ipfcacheinvalidate.so' LANGUAGE C;

CREATE TRIGGER ipfcacheinvalidate AFTER INSERT OR UPDATE OR DELETE ON validinetaddr FOR EACH ROW EXECUTE PROCEDURE ipfcacheinvalidate();

*/
