/*
//  pgipfauth - Postgres IPF Authorization Helper
//  pgipfauth
//
//  Postgres-based IPF authorization daemon.
//
//  Created by Jeffrey Frey on 2008.02.01
//  Copyright (c) 2007 University of Delaware
//
//  $Id: pgipfauth.c 122 2009-03-04 17:03:48Z frey $
*/

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <stropts.h>
#include <ctype.h>
     
#ifdef SOLARIS2
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <errno.h>
# include <syslog.h>
# include <netdb.h>
# include <sys/conf.h>
#endif

/* IP filter headers: */
#include <netinet/in.h>
#include <net/if.h>
#include <netinet/ipl.h>
#include <netinet/ip_compat.h>
#include <netinet/ip_fil.h>
#include <netinet/ip_auth.h>

/* XML headers: */
#include "libxml/tree.h"
#include "libxml/parser.h"

#include "IPFAuthCache.h"
#include "IPFDatabase.h"
#include "config.h"


/*!
  @enum Verbosity level
  
*/
enum {
  kPGIPFVerbosity_Quiet,
  kPGIPFVerbosity_Intrusive,
  kPGIPFVerbosity_Annoying
};


/*!
  @var PGIPFCLIOptions
  
  Command-line options for getopt.
*/
static struct option PGIPFCLIOptions[] =
{
  { "daemon",       no_argument,        NULL,           'd' },
  { "quiet",        no_argument,        NULL,           'q' },
  { "annoying",     no_argument,        NULL,           'a' },
  { "help",         no_argument,        NULL,           'h' },
  { "invalidator",  required_argument,  NULL,           'i' },
  { "config",       required_argument,  NULL,           'c' },
  { NULL,           0,                  NULL,            0  }
};


/*!
  @typedef PGIPFOptions
  
  Encapsulates options to the daemon itself.
*/
typedef struct {
  char              version[16];
  unsigned int      packetFlags;
  
  unsigned int      compiledPassCode,compiledBlockCode;
} PGIPFOptions;


/*!
  @enum Packet disposition flags
  
  Internal representation of the flags that we can apply to packets
  being filtered by us.  Essentially, they imply specific options be
  OR'ed into the returned packet disposition.
  
  @const kPGIPFPacketFlag_Authoritative FR_QUICK
  @const kPGIPFPacketFlag_Logging FR_LOG [| FR_LOGFIRST:if FR_PASS]
  @const kPGIPFPacketFlag_KeepState [FR_KEEPSTATE:if FR_PASS]
  @const kPGIPFPacketFlag_ReturnTCPReset [FR_RETRST:if FR_BLOCK]
*/
enum {
  kPGIPFPacketFlag_Authoritative      = 1 << 0,
  kPGIPFPacketFlag_Logging            = 1 << 1,
  kPGIPFPacketFlag_KeepState          = 1 << 2,
  kPGIPFPacketFlag_ReturnTCPReset     = 1 << 3
};


/*!
  @function PGIPFOptionsInit
  
  Initialize an options structure with default values.
*/
void
PGIPFOptionsInit(
  PGIPFOptions*     options
)
{
  strncpy(options->version,"1.0",sizeof(options->version));
  options->packetFlags = kPGIPFPacketFlag_Authoritative;
}


/*!
  @var pgIPFDefaultConfPath
  
  By default, we look for our configuration at this path.
*/
const char*       pgIPFDefaultConfPath = SYSCONFDIR"/pgipfauth.conf";


/*!
  @var pgIPFDefaultInvalidationPipePath
  
  By default, we look for our invalidation FIFO/pipe at this path.
*/
const char*       pgIPFDefaultInvalidationPipePath = SYSCONFDIR"/cache-invalidate";


/*!
  @var pgIPFSoftRestart
  
  Set to non-zero by the signal handler to trigger a soft restart of the daemon.
*/
int               pgIPFSoftRestart = 0;


/*!
  @var pgIPFShowInfo
  
  Set to non-zero by the signal handler to trigger the display of status info to stdout.
*/
int               pgIPFShowInfo = 0;


/*!
  @var pgIPFPurgeCache
  
  Set to non-zero by the signal handler to trigger a full purge of the cache.
*/
int               pgIPFPurgeCache = 0;


/*!
  @var pgIPFCacheInvalidate
  
  Set to non-zero by the signal handler to trigger an invalidation update of the cache
  (if present).
*/
int               pgIPFCacheInvalidate = 0;


/*!
  @var pgIPFSignalBrokeIPFRead
  
  Set to non-zero by the signal handler to indicate whether the ioctl() call returned
  in error due to a signal or due to a problem from IPF.
*/
int               pgIPFSignalBrokeIPFRead = 0;


/*!
  @var pgIPFTerminate
  
  Set to non-zero by the signal handler to trigger a graceful shutdown of the daemon.
*/
int               pgIPFTerminate = 0;


/*!
  @function usage
  
  Output a cli usage summary.
*/
void
usage(
  const char*     exeName
)
{
  printf("usage:\n\n"
         "  %s {options}\n\n  options:\n\n"
         "    --help/-h                 this info\n"
         "    --quiet/-q                don't print anything except critical information\n"
         "    --annoying/-a             print so much that the sysadmin will go crazy trying to read\n"
         "                              our log files\n"
         "    --daemon/-d               run as a daemon (not in the foreground)\n"
         "    --invalidator/-i [path]   use [path] as the FIFO we should watch for cache invalidation\n"
         "                              requests; default is %s\n"
         "    --config/-c [path]        use [path] as the configuration file; default configuration file\n"
         "                              is at: %s\n\n"
         "  signals:\n\n"
         "    HUP                 force the daemon to dump its cache, close the database connection,\n"
         "                        and re-read the configuration file\n"
         "    USR1                write current info for the cache and database to the daemon's\n"
         "                        stdout\n"
         "    USR2                force the daemon to purge its cache\n"
         "    TERM,ABRT,QUIT,INT  terminate the daemon gracefully\n"
         "\n",
         exeName,
         pgIPFDefaultInvalidationPipePath,
         pgIPFDefaultConfPath
    );
}


/*!
  @function log_fprintf
  
  Log-esque printing.  Prepends the output text with a timestamp and the process
  identifier.
*/
void
log_fprintf(
  FILE*       stream,
  const char* format,
  ...
)
{
  va_list     vArgs;
  time_t      now = time(NULL);
  char        timestamp[64];
  
  strftime(timestamp,64,"%c",localtime(&now));
  fprintf(stream,"%s [%d] : ",timestamp,getpid());
  va_start(vArgs,format);
  vfprintf(stream,format,vArgs);
  va_end(vArgs);
  fflush(stream);
}


/*!
  @function createPacketInfo
  
  Builds a text buffer containing the basic packet info we want to display.
  Try embedding two calls to inet_ntoa() in a printf() and you'll see why
  two printf() calls are needed -- if we're gonna make it thusly complicated,
  then encapsulate it in a function!
*/
void
createPacketInfo(
  char*     buffer,
  size_t    bufferLen,
  frauth_t* pktAuth
)
{
  int       idx;
  
  idx = snprintf(
            buffer,
            bufferLen,
            "%s:%d => ", 
            inet_ntoa(pktAuth->fra_info.fin_src),
            pktAuth->fra_info.fin_sport
          );
  if ( idx < bufferLen ) {
    idx += snprintf(
              buffer + idx,
              bufferLen - idx,
              "%s:%d",
              inet_ntoa(pktAuth->fra_info.fin_dst),
              pktAuth->fra_info.fin_dport
            );
  } else {
    idx = bufferLen - 1;
  }
  buffer[idx] = '\0';
}


/*!
  @function __pgipfauth_signals
  
  Our signal handler; very straightforward stuff, we merely set flags so that
  the main packet-processing loop will do something before again listening for
  an incoming auth request.
*/
void
__pgipfauth_signals(
  int       sig
)
{
  switch ( sig ) {
  
    case SIGHUP:
      pgIPFSoftRestart = 1;
      break;
      
    case SIGUSR1:
      pgIPFShowInfo = 1;
      break;
      
    case SIGUSR2:
      pgIPFPurgeCache = 1;
      break;
      
    case SIGIO:
      pgIPFCacheInvalidate = 1;
      break;
    
    case SIGQUIT:
    case SIGABRT:
    case SIGTERM:
    case SIGINT: {
      pgIPFTerminate = sig;
      break;
    }
    
  }
  
  /* Make sure the main processing loop doesn't claim that it broke from
     sleep because it failed to read /dev/ipfauth: */
  pgIPFSignalBrokeIPFRead = 1;
  
  /* Thanks, Solaris, make me re-register every time: */
  signal(sig , __pgipfauth_signals);
  
  /* We should already be tossed out of sleep, but why not make sure: */
  sleep(0);
}


/*!
  @function setupSignals
  
  Hookup the __pgipfauth_signals() function to the signals we're going to
  listen to.
*/
void
setupSignals()
{
  signal( SIGHUP , __pgipfauth_signals );
  signal( SIGUSR1 , __pgipfauth_signals );
  signal( SIGUSR2 , __pgipfauth_signals );
  signal( SIGIO , __pgipfauth_signals );
  signal( SIGQUIT , __pgipfauth_signals );
  signal( SIGABRT , __pgipfauth_signals );
  signal( SIGTERM , __pgipfauth_signals );
  signal( SIGINT , __pgipfauth_signals );
}


/*!
  @function xmlCharToBool
  
  Non-zero integer, "yes", "true", or "on" will result in a
  boolean true (1), everything else yields false (0).
*/
int
xmlCharToBool(
  const xmlChar*    s
)
{
  if ( s ) {
    if ( (xmlStrcasecmp(s,BAD_CAST("yes")) == 0) || 
         (xmlStrcasecmp(s,BAD_CAST("true")) == 0) ||
         (xmlStrcasecmp(s,BAD_CAST("on")) == 0)
       )
    {
      return 1;
    }
    if ( strtol((const char*)s,NULL,10) != 0 )
      return 1;
  }
  return 0;
}


/*!
  @function readConfiguration
  
  Attempt to read the XML configuration file.  Initialize a database
  and (if enabled) a cache object.  Also setup an options data structure
  using known <pgipfauth-conf> attributes.
  
  Returns non-zero if we _at least_ got a database object setup using
  the XML.  The database is the minimal piece we need to actually do
  something.
  
  The caller is reponsible for deallocating the database and cache
  objects that are returned.
*/
int
readConfiguration(
  const char*       confFilePath,
  PGIPFOptions*     newOptions,
  IPFDatabaseRef*   newDatabase,
  IPFAuthCacheRef*  newCache
)
{
  static const xmlChar*   PGIPF_xml_pgipfauthconf_version = BAD_CAST("version");
  static const xmlChar*   PGIPF_xml_pgipfauthconf_authoritative = BAD_CAST("authoritative");
  static const xmlChar*   PGIPF_xml_pgipfauthconf_logging = BAD_CAST("ipf-logging");
  static const xmlChar*   PGIPF_xml_pgipfauthconf_keepstate = BAD_CAST("ipf-keep-state");
  static const xmlChar*   PGIPF_xml_pgipfauthconf_returnrst = BAD_CAST("ipf-return-reset");
  static const xmlChar*   PGIPF_xml_yes = BAD_CAST("yes");
  
  xmlDocPtr     ourPrefDoc = NULL;
  
  /* Set default options: */
  PGIPFOptionsInit(newOptions);
  
  /* Parse the configuration XML: */
  if ( ourPrefDoc = xmlReadFile(confFilePath,NULL,0) ) {
    xmlNodePtr  rootNode = xmlDocGetRootElement(ourPrefDoc);
    
    /* Try to instantiate a database object: */
    *newDatabase = IPFDatabaseCreateWithXMLNode(rootNode->children);
    
    /* As long as we can create a database object, we can try to process
       authorizations: */
    if ( *newDatabase ) {
      xmlChar*      attrVal;
      
      /* Try to instantiate a cache object; if this fails, we're not really
         S.O.L. yet: */
      *newCache = IPFAuthCacheCreateWithXMLNode(rootNode->children);
      
      /* Grab the configure file's version number; we don't actually do
         anything with it (yet): */
      if ( attrVal = xmlGetProp(rootNode,PGIPF_xml_pgipfauthconf_version) ) {
        strncpy(newOptions->version,(const char*)attrVal,sizeof(newOptions->version));
        xmlFree(attrVal);
      }
      
      /* Try to grab the "authoritative" attribute: */
      if ( attrVal = xmlGetProp(rootNode,PGIPF_xml_pgipfauthconf_authoritative) ) {
        if ( xmlCharToBool(attrVal) )
          newOptions->packetFlags |= kPGIPFPacketFlag_Authoritative;
        else
          newOptions->packetFlags &= ~kPGIPFPacketFlag_Authoritative;
        xmlFree(attrVal);
      }
      
      /* Try to grab the "ipf-logging" attribute: */
      if ( attrVal = xmlGetProp(rootNode,PGIPF_xml_pgipfauthconf_logging) ) {
        if ( xmlCharToBool(attrVal) )
          newOptions->packetFlags |= kPGIPFPacketFlag_Logging;
        else
          newOptions->packetFlags &= ~kPGIPFPacketFlag_Logging;
        xmlFree(attrVal);
      }
      
      /* Try to grab the "ipf-keep-state" attribute: */
      if ( attrVal = xmlGetProp(rootNode,PGIPF_xml_pgipfauthconf_keepstate) ) {
        if ( xmlCharToBool(attrVal) )
          newOptions->packetFlags |= kPGIPFPacketFlag_KeepState;
        else
          newOptions->packetFlags &= ~kPGIPFPacketFlag_KeepState;
        xmlFree(attrVal);
      }
      
      /* Try to grab the "ipf-return-reset" attribute: */
      if ( attrVal = xmlGetProp(rootNode,PGIPF_xml_pgipfauthconf_returnrst) ) {
        if ( xmlCharToBool(attrVal) )
          newOptions->packetFlags |= kPGIPFPacketFlag_ReturnTCPReset;
        else
          newOptions->packetFlags &= ~kPGIPFPacketFlag_ReturnTCPReset;
        xmlFree(attrVal);
      }
    }
    
    /* Dispose of the XML doc tree, etc: */
    xmlFreeDoc(ourPrefDoc);
  }
  
  /* Build the pass and block codes that we'll return to IPF: */
  newOptions->compiledPassCode = FR_PASS | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_Authoritative ? FR_QUICK : 0 ) | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_Logging ? FR_LOG | FR_LOGFIRST : 0 ) | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_KeepState ? FR_KEEPSTATE : 0 );
  
  newOptions->compiledBlockCode = FR_BLOCK | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_Authoritative ? FR_QUICK : 0 ) | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_Logging ? FR_LOG : 0 ) | \
                                ( newOptions->packetFlags & kPGIPFPacketFlag_ReturnTCPReset ? FR_RETRST : 0 );
  
  return ( *newDatabase != NULL );
}


/*!
  @function main
  
  Duh?
*/
int
main(
  int             argc,
  const char*     argv[]
)
{
  /* We maintain two local variables for each of the configuration entities; this allows
     us to use the restart* set on HUP and only destroy and set the online* set once we
     know the restart* set is valid. */
  IPFDatabaseRef  onlineDb = NULL,restartDb = NULL;
  IPFAuthCacheRef onlineCache = NULL,restartCache = NULL;
  PGIPFOptions    onlineOpts,restartOpts;
  int             ready = 0,verbose = kPGIPFVerbosity_Intrusive,asDaemon = 0;
  int             ch,optIndex;
  const char*     exeName = argv[0];
  char*           confFile = NULL;
  char*           cacheInvldtFile = NULL;
  int             ipfauth_fd = -1;
  
  /* Cache invalidation pipeline: */
  int             invalidation_fd = -1;
  
  /* IPF data structures we'll need: */
  frauth_t        authData;
  ipfobj_t        ipfDataTmpl,ipfData;
  
  /* Initialize the ipfDataTemplate: */
  ipfDataTmpl.ipfo_rev = IPFILTER_VERSION;
  ipfDataTmpl.ipfo_size = sizeof(authData);
  ipfDataTmpl.ipfo_ptr = &authData;
  ipfDataTmpl.ipfo_type = IPFOBJ_FRAUTH;
  ipfDataTmpl.ipfo_offset = 0;
  
  /* Process those CLI options: */
  while ((ch = getopt_long(argc,(char * const *)argv,"dqahi:c:",PGIPFCLIOptions,&optIndex)) != -1) {
    switch (ch) {
      
      case 'd':
        asDaemon = 1;
        break;
        
      case 'q':
        verbose = kPGIPFVerbosity_Quiet;
        break;
      
      case 'a':
        log_fprintf(stdout,"[NOTICE] processing cli options\n");
        verbose = kPGIPFVerbosity_Annoying;
        break;
      
      case 'h':
        usage(exeName);
        exit(0);
        break;
        
      case 'i':
        if ( optarg ) {
          cacheInvldtFile = optarg;
        }
        break;
      
      case 'c':
        if ( optarg ) {
          confFile = optarg;
        }
        break;
    
    }
  }
  
  /* Use the default if we couldn't grab one on the CLI: */
  if ( confFile == NULL )
    confFile = (char*)pgIPFDefaultConfPath;
    
  /* Use the default if we couldn't grab one on the CLI: */
  if ( cacheInvldtFile == NULL )
    cacheInvldtFile = (char*)pgIPFDefaultInvalidationPipePath;
  
  /* Check for a library mismatch: */
  if ( verbose == kPGIPFVerbosity_Annoying ) {
    log_fprintf(stdout,"[NOTICE] testing for a libxml2 version mismatch\n");
  }
  LIBXML_TEST_VERSION
  
  /* Shall we run as a daemon at this point? */
  if ( asDaemon ) {
    if ( fork() != 0 ) {
      exit(0);
    }
  }
  
  /* Make sure we're ready to respond to some signals: */
  if ( verbose == kPGIPFVerbosity_Annoying ) {
    log_fprintf(stdout,"[NOTICE] setting signal handlers\n");
  }
  setupSignals();
  
  /* Open the invalidation pipeline: */
  if ( (mkfifo(cacheInvldtFile, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP) != 0) && (errno != EEXIST) ) {
    log_fprintf(
        stdout,
        "[NOTICE] could not create the cache invalidation FIFO: %s (errno = %d)\n",
        cacheInvldtFile,
        errno
      );
  } else {
    invalidation_fd = open(cacheInvldtFile, O_NONBLOCK | O_RDONLY);
    if ( invalidation_fd < 0 ) {
      log_fprintf(stdout, "[ERROR ] unable to open cache invalidation FIFO: %s (errno = %d)\n", errno);
    } else {
      /* Notify when data becomes available: */
      ioctl(invalidation_fd, I_SETSIG, S_INPUT);
    }
  }
  
  /* Setup our connection to IPF: */
	if ( (ipfauth_fd = open(IPAUTH_NAME, O_RDWR)) == -1) {
		log_fprintf(stdout,"[ERROR ] unable to open %s\n",IPAUTH_NAME);
		exit(1);
	}
	ch = 0;
	(void) ioctl(ipfauth_fd, SIOCIPFFL, &ch);
  if ( verbose > kPGIPFVerbosity_Quiet ) {
    log_fprintf(stdout,"[NOTICE] authorization flush %d\n",ch);
  }
  
  /* Try for an initial read of the configuration: */
  if ( verbose == kPGIPFVerbosity_Annoying ) {
    log_fprintf(stdout,"[NOTICE] attempting to read our configuration from %s\n",confFile);
  }
  if ( readConfiguration(confFile,&restartOpts,&restartDb,&restartCache) ) {
    onlineOpts = restartOpts;
    onlineCache = restartCache;
    onlineDb = restartDb;
    ready = 1;
    
    if ( verbose > kPGIPFVerbosity_Quiet ) {
      log_fprintf(stdout,"[NOTICE] configuration file read, daemon starting\n");
      if ( verbose == kPGIPFVerbosity_Annoying )
        pgIPFShowInfo = 1;
    }
  } else {
    log_fprintf(stdout,"[NOTICE] error in configuration file, daemon started but will\n"
                       "         not process IPF authorizations\n");
  }
  
  /* Begin processing events: */
  while ( ! pgIPFTerminate ) {
    
    /* Process the HUP, which is requesting we soft restart: */
    if ( pgIPFSoftRestart ) {
      if ( readConfiguration(confFile,&restartOpts,&restartDb,&restartCache) ) {
        
        /* Successfully read the configuration and ready to "restart": */
        if ( onlineDb ) {
          IPFDatabaseRelease(onlineDb);
          onlineDb = NULL;
        }
        if ( onlineCache ) {
          IPFAuthCacheRelease(onlineCache);
          onlineCache = NULL;
        }
        
        onlineOpts = restartOpts;
        onlineCache = restartCache;
        onlineDb = restartDb;
        
        ready = 1;
        
        if ( verbose > kPGIPFVerbosity_Quiet ) {
          log_fprintf(stdout,"[NOTICE] configuration file read, daemon reset\n");
          if ( verbose == kPGIPFVerbosity_Annoying )
            pgIPFShowInfo = 1;
        }
        
      } else {
        log_fprintf(stdout,"[NOTICE] error while reading configuration file, daemon continuing with\n"
                           "         previous configuration parameters\n");
      }
      
      pgIPFSoftRestart = 0;
    }
    
    /* Cache purge? */
    if ( pgIPFPurgeCache ) {
      if ( onlineCache ) {
        IPFAuthCachePurgeAll(onlineCache);
        if ( verbose > kPGIPFVerbosity_Quiet )
          log_fprintf(stdout,"[NOTICE] the internal authorization cache has been purged\n");
      }
      pgIPFPurgeCache = 0;
    }
    
    /* Check for any cache invalidations: */
    if ( pgIPFCacheInvalidate ) {
      if ( onlineCache ) {
        static char*      invalidateBuffer = NULL;
        static size_t     invalidateBufferLen = 0;
        int               gulp, exitLoop = 0;
        
        /* Initial allocation of our pervasive character buffer: */
        if ( invalidateBuffer == NULL ) {
          if ( (invalidateBuffer = malloc(64)) ) {
            invalidateBufferLen = 64;
          }
        }
        
        /* While we have a character buffer and the loop is running... */
        while ( invalidateBuffer && ! exitLoop ) {
          int       idx = 0;
          size_t    count;
          char      c;
          
          /* Read characters until we hit whitespace (which signals the separation between
             individual IPs).  For each chunk of text, convert to a binary IP form and
             ask the cache to invalidate it. */
          while ( (count = read(invalidation_fd, &c, 1)) > 0 ) {
            if ( isspace(c) ) {
              invalidateBuffer[idx] = '\0';
              break;
            }
            invalidateBuffer[idx++] = c;
            if ( idx + 1 == invalidateBufferLen ) {
              char* grownBuffer = realloc(invalidateBuffer, invalidateBufferLen + 64);
              
              /* Grow our pervasive character buffer to accomodate the latest line we're
                 attempting to read: */
              if ( grownBuffer ) {
                invalidateBuffer = grownBuffer;
                invalidateBufferLen += 64;
              } else {
                exitLoop = 1;
                break;
              }
            }
          }
          
          /* Nothing was read, time to get out of Dodge: */
          if ( count == 0 && idx == 0 )
            exitLoop = 1;
          
          /* Do we have a string to process? */
          if ( ! exitLoop && idx ) {
            struct in_addr    ipAddress;
            
            /* Convert to a binary IP address: */
            if ( inet_aton(invalidateBuffer, &ipAddress) ) {
              /* Success -- invalidate the sucker: */
              if ( verbose > kPGIPFVerbosity_Quiet )
                log_fprintf(stdout, "[NOTICE] invalidating %s in cache\n", invalidateBuffer);
              IPFAuthCacheInvalidateIP(onlineCache, &ipAddress);
            } else {
              /* Not an IP address, doofus: */
              if ( verbose > kPGIPFVerbosity_Quiet )
                log_fprintf(stdout, "[ERROR ] invalid IP address on invalidation pipe: %s\n", invalidateBuffer);
            }
          }
        }
        /* Dump the rest of the data -- mainly useful in case we had a memory allocation error, etc: */
        while ( read(invalidation_fd, &gulp, sizeof(gulp)) > 0 );
      } else {
        int     gulp;
        
        /* Dump everything, we don't have a cache! */
        while ( read(invalidation_fd, &gulp, sizeof(gulp)) > 0 );
      }
      pgIPFCacheInvalidate = 0;
    }
    
    /* Display status information for the database, cache, etc: */
    if ( pgIPFShowInfo ) {
      
      log_fprintf(stdout,"[INFO  ] configuration version: %s\n",onlineOpts.version);
      log_fprintf(stdout,"[INFO  ] configuration options:\n");
      log_fprintf(stdout,"[INFO  ]   [%c] daemon is authoritative?\n",
              ( onlineOpts.packetFlags & kPGIPFPacketFlag_Authoritative ? 'X' : ' ' )
          );
      log_fprintf(stdout,"[INFO  ]   [%c] (ipf) log packet info?\n",
              ( onlineOpts.packetFlags & kPGIPFPacketFlag_Logging ? 'X' : ' ' )
          );
      log_fprintf(stdout,"[INFO  ]   [%c] (ipf) keep packet state?\n",
              ( onlineOpts.packetFlags & kPGIPFPacketFlag_KeepState ? 'X' : ' ' )
          );
      log_fprintf(stdout,"[INFO  ]   [%c] (ipf) send TCP RESET for blocked packets?\n",
              ( onlineOpts.packetFlags & kPGIPFPacketFlag_KeepState ? 'X' : ' ' )
          );
      
      if ( onlineDb ) {
        log_fprintf(stdout,"[INFO  ] database configuration:\n");
        IPFDatabaseInfo(onlineDb);
      } else {
        log_fprintf(stdout,"[INFO  ] no database configured -- we're not doing anything.\n");
      }
      
      if ( onlineCache ) {
        log_fprintf(stdout,"[INFO  ] authorization cache configuration:\n");
        IPFAuthCacheInfo(onlineCache);
      } else {
        log_fprintf(stdout,"[INFO  ] authorization caching is not enabled.\n");
      }
      
      pgIPFShowInfo = 0;
    }
    fflush(stdout);
    
    /* If we're ready to go, then check for a packet we need to analyze: */
    if ( ready ) {
      int       wasCached = 0;
      
      /* We don't need no stinking packet data: */
      memset(&authData, 0, sizeof(authData));
      /*authData.fra_buf = NULL;
      authData.fra_len = 0;*/
      
      ipfData = ipfDataTmpl;
      
      if ( ioctl(ipfauth_fd, SIOCAUTHW, &ipfData) == 0 ) {
        IPFAuthDisposition    packetDisposition = kIPFAuthDisposition_Block;
        
        /* If the cache is available, use it: */
        if ( onlineCache ) {
          /* Expunge expired cache entries: */
          IPFAuthCachePurgeExpired(onlineCache);
          
          /* Check if it's still in the cache: */
          if ( ! IPFAuthCacheIsAuthorized(onlineCache, &authData.fra_info.fin_src, authData.fra_info.fin_dport, &packetDisposition) ) {
            /* Try to authorize via the database: */
            packetDisposition = IPFDatabaseAuthorizeIP(
                                    onlineDb,
                                    &authData.fra_info.fin_src,
                                    authData.fra_info.fin_dport
                                  );
            if ( packetDisposition != kIPFAuthDisposition_Unknown ) {
              /* Cache our discovered authorization state: */
              IPFAuthCacheAuthorize(onlineCache, &authData.fra_info.fin_src, authData.fra_info.fin_dport, packetDisposition);
            }
          } else {
            wasCached = 1;
          }
        } else {
          /* Try to authorize via the database: */
          packetDisposition = IPFDatabaseAuthorizeIP(onlineDb, &authData.fra_info.fin_src, authData.fra_info.fin_dport);
        }
        
        /* Any IP that yields an unknown disposition means there was a database
           error of some sort; we don't cache this disposition, and respond by
           always printing an error message: */
        if ( packetDisposition == kIPFAuthDisposition_Unknown ) {
          char      pktInfo[64];
          
          createPacketInfo(pktInfo, 64, &authData);
          log_fprintf(
              stdout,
              "[ERROR ] database module returned unknown disposition for %s\n",
              pktInfo
            );
        }
        
        /* Set the IPF version of packet disposition: */
        authData.fra_pass = ( packetDisposition == kIPFAuthDisposition_Pass ? \
                              onlineOpts.compiledPassCode : \
                              onlineOpts.compiledBlockCode );
        
        /* If verbosity is set to annoying, display packet state: */
        if ( verbose == kPGIPFVerbosity_Annoying ) {
          char      pktInfo[64];
          
          createPacketInfo(pktInfo, 64, &authData);
          log_fprintf(stdout,"[PACKET] %-44s %s (%08X)%s\n",
              pktInfo,
              ( packetDisposition == kIPFAuthDisposition_Pass ? "PASS " : "BLOCK" ),
              authData.fra_pass,
              ( wasCached ? "  [CACHE HIT]" : "" )
            );
        }
        
        /* Try handing the altered packet disposition back to IPF: */
        if ( ioctl(ipfauth_fd, SIOCAUTHR, &ipfData) != 0 ) {
          log_fprintf(stdout,"[ERROR ] ipf was unable to read the packet state we determined\n");
        }
      } else if ( ! pgIPFSignalBrokeIPFRead ) {
        log_fprintf(stdout,"[ERROR ] unable to retrieve the packet information from ipf\n");
      } else {
        pgIPFSignalBrokeIPFRead = 0;
      }
    }
    
  }
  
  if ( verbose > kPGIPFVerbosity_Quiet ) {
    if ( verbose == kPGIPFVerbosity_Annoying ) {
      log_fprintf(stdout,"[NOTICE] final status information for database and cache:\n");
      
      if ( onlineDb ) {
        log_fprintf(stdout,"[INFO  ] database configuration:\n");
        IPFDatabaseInfo(onlineDb);
      } else {
        log_fprintf(stdout,"[INFO  ] no database configured -- we're not doing anything.\n");
      }
      
      if ( onlineCache ) {
        log_fprintf(stdout,"[INFO  ] authorization cache configuration:\n");
        IPFAuthCacheInfo(onlineCache);
      } else {
        log_fprintf(stdout,"[INFO  ] authorization caching is not enabled.\n");
      }
    }
    log_fprintf(stdout,"[NOTICE] daemon terminating on signal: %d\n",pgIPFTerminate);
  }
  
  if ( invalidation_fd >= 0 )
    close(invalidation_fd);
    
  if ( onlineDb )
    IPFDatabaseRelease(onlineDb);
  
  if ( onlineCache )
    IPFAuthCacheRelease(onlineCache);
  
  return 0;
}
